# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the ORS directions API."""

import json

from openrouteservice import convert, validator

def directions(client,
               coordinates,
               profile='driving-car',
               format_out=None,
               preference=None,
               units=None,
               language=None,
               geometry=None,
               geometry_format=None,
               geometry_simplify=None,
               instructions=None,
               instructions_format=None,
               roundabout_exits=None,
               attributes=None,
               # maneuvers=None,
               radiuses=None,
               bearings=None,
               continue_straight=None,
               elevation=None,
               extra_info=None,
               optimized=True,
               options=None,
               dry_run=None):
    """Get directions between an origin point and a destination point.

    For more information, visit https://go.openrouteservice.org/documentation/.

    :param coordinates: The coordinates tuple the route should be calculated
        from. In order of visit.
    :type origin: list or tuple of coordinate lists or tuples

    :param profile: Specifies the mode of transport to use when calculating
        directions. One of ["driving-car", "driving-hgv", "foot-walking",
        "foot-hiking", "cycling-regular", "cycling-road",requests_kwargs
        "cycling-safe", "cycling-mountain", "cycling-tour",
        "cycling-electric",]. Default "driving-car".
    :type mode: string

    :param preference: Specifies the routing preference. One of ["fastest, "shortest",
        "recommended"]. Default "fastest".
    :type preference: string

    :param units: Specifies the distance unit. One of ["m", "km", "mi"]. Default "m".
    :type units: string

    :param language: Language for routing instructions. One of ["en", "de", "cn",
        "es", "ru", "dk", "fr", "it", "nl", "br", "se", "tr", "gr"]. Default "en".
    :type language: string

    :param language: The language in which to return results.
    :type language: string

    :param geometry: Specifies whether geometry should be returned. Default True.
    :type geometry: boolean

    :param geometry_format: Specifies which geometry format should be returned.
        One of ["encodedpolyline", "geojson", "polyline"]. Default: "encodedpolyline".
    :type geometry_format: string

    :param geometry_simplify: Specifies whether to simplify the geometry.
        Default False.
    :type geometry_simplify: boolean

    :param instructions: Specifies whether to return turn-by-turn instructions.
        Default True.
    :type instructions: boolean

    :param instructions_format: Specifies the the output format for instructions.
        One of ["text", "html"]. Default "text".
    :type instructions_format: string

    :param roundabout_exits: Provides bearings of the entrance and all passed
        roundabout exits. Adds the 'exit_bearings' array to the 'step' object
        in the response. Default False.
    :type roundabout_exits: boolean

    :param attributes: Returns route attributes on ["avgspeed", "detourfactor", "percentage"].
        Must be a list of strings. Default None.
    :type attributes: list or tuple of strings

    :param radiuses: A list of maximum distances (measured in
        meters) that limit the search of nearby road segments to every given waypoint.
        The values must be greater than 0, the value of -1 specifies no limit in
        the search. The number of radiuses must correspond to the number of waypoints.
        Default None.
    :type radiuses: list or tuple

    :param bearings: Specifies a list of pairs (bearings and
        deviations) to filter the segments of the road network a waypoint can
        snap to. For example bearings=[[45,10],[120,20]]. Each pair is a
        comma-separated list that can consist of one or two float values, where
        the first value is the bearing and the second one is the allowed deviation
        from the bearing. The bearing can take values between 0 and 360 clockwise
        from true north. If the deviation is not set, then the default value of
        100 degrees is used. The number of pairs must correspond to the number
        of waypoints. Setting optimized=false is mandatory for this feature to
        work for all profiles. The number of bearings corresponds to the length
        of waypoints-1 or waypoints. If the bearing information for the last waypoint
        is given, then this will control the sector from which the destination
        waypoint may be reached.
    :type bearings: list or tuple or lists or tuples

    :param continue_straight: Forces the route to keep going straight at waypoints
        restricting u-turns even if u-turns would be faster. This setting
        will work for all profiles except for driving-*. In this case you will
        have to set optimized=false for it to work. True or False. Default False.
    :type continue_straight: boolean

    :param elevation: Specifies whether to return elevation values for points.
        Default False.
    :type elevation: boolean

    :param extra_info: Returns additional information on ["steepness", "suitability",
        "surface", "waycategory", "waytype", "tollways", "traildifficulty"].
        Must be a list of strings. Default None.
    :type extra_info: list or tuple of strings

    :param optimized: If set True, uses Contraction Hierarchies.
        Default True.
    :type optimized: boolean

    :param options: Refer to https://go.openrouteservice.org/documentation for
        detailed documentation. Construct your own dict() following the example
        of the minified options object. Will be converted to json automatically.
    :type options: dict
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean

    :raises ValueError: When parameter has wrong value.
    :raises TypeError: When parameter is of wrong type.

    :returns: sanitized set of pararmeters
    :rtype: call to Client.request()
    """
    
    validator.validator(locals(), 'directions')

    params = {
        "coordinates": convert._build_coords(coordinates)
        }

    if profile:
        # NOTE(broady): the mode parameter is not validated by the Maps API
        # server. Check here to prevent silent failures.
        params["profile"] = profile

    if format_out:
        params["format"] = format_out

    if preference:
        params["preference"] = preference

    if units:
        params["units"] = units

    if language:
        params["language"] = language

    if geometry:
        # not checked on backend, check here
        params["geometry"] = convert._convert_bool(geometry)

    if geometry_format:
        params["geometry_format"] = geometry_format
        
    if geometry_simplify:
        # not checked on backend, check here
        if extra_info:
            params["geometry_simplify"] = 'false'
        else:
            params["geometry_simplify"] = convert._convert_bool(geometry_simplify)
        
    if instructions:
        # not checked on backend, check here
        params["instructions"] = convert._convert_bool(instructions)

    if instructions_format:
        params["instructions_format"] = instructions_format

    if roundabout_exits:
        # not checked on backend, check here
        params["roundabout_exits"] = convert._convert_bool(roundabout_exits)

    if attributes:
        # not checked on backend, check here
        params["attributes"] = convert._pipe_list(attributes)
        
    if radiuses:
        params["radiuses"] = convert._pipe_list(radiuses)
        
    if bearings:
        params["bearings"] = convert._pipe_list([convert._comma_list(pair) for pair in bearings])
        
    if continue_straight:
        # not checked on backend, check here
        params["continue_straight"] = convert._convert_bool(continue_straight)

    if elevation:
        # not checked on backend, check here
        params["elevation"] = convert._convert_bool(elevation)

    if extra_info:
        # not checked on backend, check here
        params["extra_info"] = convert._pipe_list(extra_info)

    if optimized:
        # not checked on backend, check here
        if bearings or continue_straight in (True, 'true'):
            params["optimized"] = 'false'
            print("Set optimized='false' due to incompatible parameter settings.")
        else:
            params["optimized"] = convert._convert_bool(optimized)
    
    if options:
        params['options'] = json.dumps(options)

    return client.request("/directions", params, dry_run=dry_run)


# -*- coding: utf-8 -*-

PARAM_POINT = [8.34234, 48.23424]
PARAM_LINE = [(8.34234, 48.23424), (8.34423, 48.26424)]
PARAM_POLY = [[(8.34234, 48.23424), (8.34423, 48.26424)]]

PARAM_INT_BIG = 500
PARAM_INT_SMALL = 50
PARAM_LIST_ONE = [PARAM_INT_SMALL, PARAM_INT_SMALL]
PARAM_LIST_TWO = [PARAM_LIST_ONE, PARAM_LIST_ONE]

PARAM_GEOJSON_POINT = {'type': 'Point', 'coordinates': PARAM_POINT}
PARAM_GEOJSON_LINE = {'type': 'LineString', 'coordinates': PARAM_LINE}
PARAM_GEOJSON_POLY = {'type': 'Polygon', 'coordinates': PARAM_POLY}

ENDPOINT_DICT = {'directions': {'coordinates': PARAM_LINE,
                                'profile': 'foot-walking',
                                'preference': 'fastest',
                                'units': 'mi',
                                'language': 'en',
                                'geometry': 'true',
                                'geometry_format': 'geojson',
                                'geometry_simplify': 'false',
                                'instructions': 'false',
                                'instructions_format': 'html',
                                'roundabout_exits': 'true',
                                'attributes': ['avgspeed'],
                                'radiuses': PARAM_LIST_ONE,
                                'bearings': PARAM_LIST_TWO,
                                'continue_straight': 'true',
                                'elevation': 'true',
                                'extra_info': ['steepness', 'suitability'],
                                'optimized': 'false',
                                'options': {'maximum_speed': PARAM_INT_SMALL}
                            },
                'isochrones': {'locations': PARAM_LINE,
                                 'profile': 'cycling-regular',
                                 'range_type': 'distance',
                                 'intervals': [PARAM_INT_BIG],
                                 'units': 'm',
                                 'location_type': 'destination',
                                 'attributes': ['area', 'reachfactor'],
                                 'segments': [PARAM_INT_SMALL]
                             },
                'distance_matrix': {'locations': PARAM_LINE,
                                     'sources': [1],
                                     'destinations': 'all',
                                     'profile': 'cycling-regular',
                                     'metrics': ['duration', 'distance'],
                                     'resolve_locations': 'true',
                                     'units': 'mi',
                                     'optimized': 'false'
                                    },
                'elevation_point': {'format_in': 'geojson',
                                     'format_out': 'point',
                                     'geometry': PARAM_GEOJSON_POINT,
                                     'dataset': 'srtm'
                                     },
                'elevation_line': {'format_in': 'geojson',
                                     'format_out': 'polyline',
                                     'geometry': PARAM_GEOJSON_LINE,
                                     'dataset': 'srtm'
                                     },
                'pelias_search': {'text': 'Heidelberg',
                                   'focus_point': PARAM_POINT,
                                   'rect_min_x': PARAM_INT_BIG,
                                   'rect_min_y': PARAM_INT_BIG,
                                   'rect_max_x': PARAM_INT_BIG,
                                   'rect_max_y': PARAM_INT_BIG,
                                   'circle_point': PARAM_POINT,
                                   'circle_radius': PARAM_INT_SMALL,
                                   'sources': ['osm', 'wof', 'gn'],
                                   'layers': ['locality', 'county', 'region'],
                                   'country': 'de',
                                   'size': PARAM_INT_SMALL,
                                   },
                'pelias_autocomplete': {'text': 'Heidelberg',
                                   'focus_point': PARAM_POINT,
                                   'rect_min_x': PARAM_INT_BIG,
                                   'rect_min_y': PARAM_INT_BIG,
                                   'rect_max_x': PARAM_INT_BIG,
                                   'rect_max_y': PARAM_INT_BIG,
                                   'sources': ['osm', 'wof', 'gn'],
                                   'layers': ['locality', 'county', 'region'],
                                   'country': 'de',
                                   },
                'pelias_structured':{'address': 'Berliner Straße 45',
                                       'neighbourhood': 'Neuenheimer Feld',
                                       'borough': 'Heidelberg',
                                       'locality': 'Heidelberg',
                                       'county': 'Rhein-Neckar-Kreis',
                                       'region': 'Baden-Württemberg',
                                       'postalcode': 69120,
                                       'country': 'de',
                                       },
                'pelias_reverse': {'point': PARAM_POINT,
                                    'circle_radius': PARAM_INT_SMALL,
                                    'sources': ['osm', 'wof', 'gn'],
                                    'layers': ['locality', 'county', 'region'],
                                    'country': 'de',
                                    'size': PARAM_INT_SMALL,
                                    },
                'pois': {'request': 'pois',
                         'geojson': PARAM_GEOJSON_POINT,
                         'bbox': PARAM_LINE,
                         'buffer': PARAM_INT_SMALL,
                         'filter_category_ids': [PARAM_INT_SMALL],
                         'filter_category_group_ids': [PARAM_INT_BIG],
                         'filters_custom': {'name': 'Deli',
                                            'wheelchair': ['yes', 'limited'],
                                            'smoking': ['dedicated', 'separated'],
                                            'fee': ['yes', 'no']},
                         'limit': PARAM_INT_SMALL,
                         'sortby': 'distance',
                         }    
                }


# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the convert module."""

import unittest

from openrouteservice import convert


class ConvertTest(unittest.TestCase):

    def test_build_single_coord_tuple(self):
        expected = "1,2"
        ll = (1, 2)
        self.assertEqual(expected, convert._build_coords(ll))

        ll = [1, 2]
        self.assertEqual(expected, convert._build_coords(ll))

        with self.assertRaises(TypeError):
            convert._build_coords(1)

        with self.assertRaises(TypeError):
            convert._build_coords({'lat':1, 'lon':2})
            

        with self.assertRaises(TypeError):
            convert._build_coords('1,2')
            
            
    def test_build_multi_coord_tuple(self):
        expected = "1,2|3,4"

        ll = ((1,2),(3,4))
        self.assertEqual(expected, convert._build_coords(ll))

        ll = [(1,2),(3,4)]
        self.assertEqual(expected, convert._build_coords(ll))
        
        ll = ([1,2],[3,4])
        self.assertEqual(expected, convert._build_coords(ll))
        
        ll = [[1,2],[3,4]]
        self.assertEqual(expected, convert._build_coords(ll))
        
        with self.assertRaises(TypeError):
            convert._build_coords({{'lat':1, 'lon':2},{'lat':3, 'lon':4}})
            

        with self.assertRaises(TypeError):
            convert._build_coords('[1,2],[3,4]')
            
    
    def test_convert_bool(self):
        self.assertEqual('true', convert._convert_bool('True'))
        self.assertEqual('true', convert._convert_bool('true'))
        self.assertEqual('true', convert._convert_bool(True))
        
    
    def test_polyline_decode_3d(self):
        syd_mel_route = (r"mlqlHat`t@OiACMvAs@HCPGJ?JAJBRFTRLJPNHDNDJ"
                         "@D?fACRAZCPAb@AF?HAfBQJEDAn@QFC@QD_@@QFe@Bg"
                         "@@KBy@?M@a@@q@?iE?C?OGgAkEwUQ{@c@gBQeAYeCIe"
                         "AWmDAIImACUOyBIeAC}@Ey@?QLC@_@@KBiAVmDF]Ni@"
                         "Zu@RYBA^_@~A{A`Ai@JCPGf@Qf@]X_@BMAMIKuBTI?G"
                         "E?A?ADOnCsB\c@DGDIl@sAJUFMBGJUP[DCD@DP@l@?R"
                         "?h@Bx@PnAAl@?BAFc@rAAB?@BRHBFEN[FQFQRg@Rw@J"
                         "g@Ny@DUDOJe@N_ADm@BkBGcC@s@Du@l@eEZgBP_AHe@"
                         "He@Fc@RuATaA?SCWAGIOQS[Qu@Ym@C}@R{@`@m@p@Wj"
                         "@]nAGBE?KGAE?E?KVcB`@eB^mAn@uALUJSj@y@fA}@f"
                         "@k@BGHM^k@r@qAHSLU^i@bA_Af@q@PYFKHIHCJ?RLFN"
                         "XjAj@tDj@rERzBLzCHp@xAdKLf@RXTDNEBCFGDEDE@G"
                         "@GDKBGRc@Xi@N[JUf@u@l@o@f@c@h@]XMfQ}D|EcAlA"
                         "ORIJQ?C?CAUKOSGwAMa@M_EsBcBqA_A{@k@q@sCcEi@"
                         "gAWo@[gAYyAMy@y@aNMyAc@uDS_As@uBMc@Ig@SeBKc"
                         "@Uy@AI@A]GGCMIiCmAGCWMqAk@")

        points = convert.decode_polyline(syd_mel_route, True)['coordinates']
        self.assertEqual(len(points[0]), 3)
        self.assertAlmostEqual(8.69201, points[0][0], places=5)
        self.assertAlmostEqual(49.410151, points[0][1], places=5)
        self.assertAlmostEqual(0.1, points[0][2], places=2)
        self.assertAlmostEqual(8.69917, points[-1][0], places=5) 
        self.assertAlmostEqual(49.41868 , points[-1][1], places=5)
        self.assertAlmostEqual(12.5, points[-1][2], places=2)
        
    
    def test_polyline_decode_2d(self):
        syd_mel_route = (r"u`rgFswjpAKD")

        points = convert.decode_polyline(syd_mel_route, False)['coordinates']
        self.assertEqual(len(points[0]), 2)
        self.assertAlmostEqual([13.3313, 38.10843], points[0], places=5)
        self.assertAlmostEqual([13.33127, 38.10849], points[1], places=5)


# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Converts Python types to string representations suitable for ORS API server.
"""

def _pipe_list(arg):
    """Convert list of values to pipe-delimited string"""
    if not _is_list(arg):
        raise TypeError(
            "Expected a list or tuple, "
            "but got {}".format(type(arg).__name__))
    return "|".join(map(str,arg))


def _comma_list(arg):
    """Convert list to comma-separated string"""
    if not _is_list(arg):
        raise TypeError(
            "Expected a list or tuple, "
            "but got {}".format(type(arg).__name__))
    return ",".join(map(str,arg))


def _convert_bool(boolean):
    """Convert to stringified boolean"""
        
    return str(boolean).lower()


def _format_float(arg):
    """Formats a float value to be as short as possible.

    Trims extraneous trailing zeros and period to give API
    args the best possible chance of fitting within 2000 char
    URL length restrictions.

    For example:

    format_float(40) -> "40"
    format_float(40.0) -> "40"
    format_float(40.1) -> "40.1"
    format_float(40.001) -> "40.001"
    format_float(40.0010) -> "40.001"

    :param arg: The lat or lng float.
    :type arg: float

    :rtype: string
    """
    return ("{}".format(round(float(arg), 6)).rstrip("0").rstrip("."))


def _build_coords(arg):
    """Converts one or many lng/lat pair(s) to a comma-separated, pipe 
    delimited string. Coordinates will be rounded to 5 digits.

    For example:

    convert.build_coords([(151.2069902,-33.8674869),(2.352315,48.513158)])
    # '151.20699,-33.86749|2.35232,48.51316'

    :param arg: The lat/lon pair(s).
    :type arg: list or tuple
    
    :rtype: str
    """
    if _is_list(arg):
        return _pipe_list(_concat_coords(arg))
    else:
        raise TypeError(
            "Expected a list or tuple of lng/lat tuples or lists, "
            "but got {}".format(type(arg).__name__))


def _concat_coords(arg):
    """Turn the passed coordinate tuple(s) in comma separated coordinate tuple(s).
    
    :param arg: coordinate pair(s)
    :type arg: list or tuple
    
    :rtype: list of strings
    """
    if all(_is_list(tup) for tup in arg):
        # Check if arg is a list/tuple of lists/tuples
        return [_comma_list(map(_format_float, tup)) for tup in arg]
    else:
        return [_comma_list(_format_float(coord) for coord in arg)]


def _is_list(arg):
    """Checks if arg is list-like."""    
    if isinstance(arg, dict):
        return False
    if isinstance(arg, str): # Python 3-only, as str has __iter__
        return False
    return (not _has_method(arg, "strip")
            and _has_method(arg, "__getitem__")
            or _has_method(arg, "__iter__"))


def _has_method(arg, method):
    """Returns true if the given object has a method with the given name.

    :param arg: the object

    :param method: the method name
    :type method: string

    :rtype: bool
    """
    return hasattr(arg, method) and callable(getattr(arg, method))


def decode_polyline(polyline, is3d=False):
    """Decodes a Polyline string into a GeoJSON structure.
    
    :param polyline: An encoded polyline, only the geometry.
    :type polyline: string
    
    :param is3d: Specifies if geometry contains Z component.
    :type is3d: boolean
    
    :returns: GeoJSON Linestring geometry
    :rtype: dict
    """
    points = []
    index = lat = lng = z= 0

    while index < len(polyline):
        result = 1
        shift = 0
        while True:
            b = ord(polyline[index]) - 63 - 1
            index += 1
            result += b << shift
            shift += 5
            if b < 0x1f:
                break
        lat += (~result >> 1) if (result & 1) != 0 else (result >> 1)

        result = 1
        shift = 0
        while True:
            b = ord(polyline[index]) - 63 - 1
            index += 1
            result += b << shift
            shift += 5
            if b < 0x1f:
                break
        lng += ~(result >> 1) if (result & 1) != 0 else (result >> 1)
        
        if is3d:
            result = 1
            shift = 0
            while True:
                b = ord(polyline[index]) - 63 - 1
                index += 1
                result += b << shift
                shift += 5
                if b < 0x1f:
                    break
            if (result & 1) != 0:
                z += ~(result >> 1)
            else:
                z += (result >> 1) 
                
            points.append([round(lng * 1e-5, 6), round(lat * 1e-5, 6), round(z*1e-2,1)])   
            
        else:
            points.append([round(lng * 1e-5, 6), round(lat * 1e-5, 6)])
    
    geojson = {u'type': u'LineString', u'coordinates': points}

    return geojson



# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.


"""Tests for the Pelias geocoding module."""

import responses

import test as _test
import unittest
import openrouteservice
from collections import OrderedDict


class GeocodingPeliasTest(_test.TestCase):
    def setUp(self):
        self.key = 'sample_key'
        self.client = openrouteservice.Client(self.key)
        self.search = {'text': 'Heidelberg',
                       'focus_point': (8.675786, 49.418431),
                       'rect_min_x': 8.573179,
                       'rect_min_y': 49.351764,
                       'rect_max_x': 8.79405,
                       'rect_max_y': 49.459693,
                       'circle_point': (8.675786, 49.418431),
                       'circle_radius': 50,
                       'sources': ['osm', 'wof', 'gn'],
                       'layers': ['locality', 'county', 'region'],
                       'country': 'de',
                       'size': 5,
                       }
        self.autocomplete = {'text': 'Heidelberg',
                             'focus_point': (8.675786, 49.418431),
                             'rect_min_x': 8.573179,
                             'rect_min_y': 49.351764,
                             'rect_max_x': 8.79405,
                             'rect_max_y': 49.459693,
                             'sources': ['osm', 'wof', 'gn'],
                             'layers': ['locality', 'county', 'region'],
                             'country': 'de',
                             }
        self.structured = {'address': 'Berliner Straße 45',
                           'neighbourhood': 'Neuenheimer Feld',
                           'borough': 'Heidelberg',
                           'locality': 'Heidelberg',
                           'county': 'Rhein-Neckar-Kreis',
                           'region': 'Baden-Württemberg',
                           'postalcode': 69120,
                           'country': 'de',
                           }
        self.reverse = {'point': (8.675786, 49.418431),
                        'circle_radius': 50,
                        'sources': ['osm', 'wof', 'gn'],
                        'layers': ['locality', 'county', 'region'],
                        'country': 'de',
                        'size': 5,
                        }

    @responses.activate
    def test_full_search(self):
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/geocode/search',
                      body='{"status":"OK","results":[]}',
                      status=200,
                      content_type='application/json')

        results = self.client.pelias_search(**self.search)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual(
            'https://api.openrouteservice.org/geocode/search?boundary.circle.lat=49.418431&boundary.circle.lon=8.675786&boundary.circle.radius=50&boundary.rect.max_lon=8.79405&boundary.rect.max_lat=49.459693&boundary.rect.min_lat=49.351764&boundary.rect.min_lon=8.573179&country=de&focus.point.lat=49.418431&focus.point.lon=8.675786&layers=locality%2Ccounty%2Cregion&size=5&sources=osm%2Cwof%2Cgn&text=Heidelberg&api_key=sample_key'.format(
                self.key),
            responses.calls[0].request.url)

    @responses.activate
    def test_full_autocomplete(self):
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/geocode/autocomplete',
                      body='{"status":"OK","results":[]}',
                      status=200,
                      content_type='application/json')

        results = self.client.pelias_autocomplete(**self.autocomplete)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual(
            'https://api.openrouteservice.org/geocode/autocomplete?boundary.rect.max_lon%09=49.459693&boundary.rect.min_lat%09=49.351764&boundary.rect.min_lon%09=8.573179&boundary.country=de&focus.point.lat=49.418431&focus.point.lon=8.675786&layers=locality%2Ccounty%2Cregion&sources=osm%2Cwof%2Cgn&text=Heidelberg&api_key=sample_key'.format(
                self.key),
            responses.calls[0].request.url)

    @responses.activate
    def test_full_structured(self):
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/geocode/search/structured',
                      body='{"status":"OK","results":[]}',
                      status=200,
                      content_type='application/json')

        results = self.client.pelias_structured(**self.structured)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual(
            'https://api.openrouteservice.org/geocode/search/structured?address=Berliner%20Straße%2045&neighbourhood=Neuenheimer%20Feld&borough=Heidelberg&locality=Heidelberg&county=Rhein-Neckar-Kreis&region=Baden-Württemberg&postalcode=69120&country=de&api_key=sample_key'.format(
                self.key),
            responses.calls[0].request.url)

    @responses.activate
    def test_full_reverse(self):
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/geocode/reverse',
                      body='{"status":"OK","results":[]}',
                      status=200,
                      content_type='application/json')

        results = self.client.pelias_reverse(**self.reverse)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual(
            'https://api.openrouteservice.org/geocode/reverse?boundary.circle.radius=50&country=de&layers=locality%2Ccounty%2Cregion&point.lat=49.418431&point.lon=8.675786&size=5&sources=osm%2Cwof%2Cgn&api_key=sample_key'.format(
                self.key),
            responses.calls[0].request.url)



# -*- coding: utf-8 -*-

# Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Performs requests to the ORS Places API."""

from openrouteservice import convert, validator

def places(client, request,
                geojson=None,
                bbox=None,
                buffer=None,
                filter_category_ids=None,
                filter_category_group_ids=None,
                filters_custom=None,
                limit=None,
                sortby=None,
                dry_run=None
                    ):
    """ Gets POI's filtered by specified parameters.

    :param request: Type of request. One of ['pois', 'list', 'stats'].
        'pois': returns geojson of pois;
        'stats': returns statistics of passed categories;
        'list': returns mapping of category ID's to textual representation.
    :type request: string

    :param geojson: GeoJSON dict used for the query.
    :type geojson: dict

    :param buffer: Buffers geometry of 'geojson' or 'bbox' with the specified
        value in meters.
    :type intervals: integer

    :param filter_category_ids: Filter by ORS custom category IDs. See
        https://github.com/GIScience/openrouteservice-docs#places-response
        for the mappings.
    :type units: list of integer

    :param filter_category_group_ids: Filter by ORS custom high-level category
        groups. See
        https://github.com/GIScience/openrouteservice-docs#places-response
        for the mappings.        
    :type attributes: list of integer

    :param filters_custom: Specify additional filters by key/value. Default ORS
        filters are
        'name': free text
        'wheelchair': ['yes', 'limited', 'no', 'designated']
        'smoking': ['dedicated','yes','separated','isolated', 'no', 'outside']
        'fee': ['yes','no', 'str']
    :type filters_custom: dict of lists/str

    :param limit: limit for POI queries.
    :type limit: integer
                       base_url='http://localhost:5000'
    
    :param sortby: Sorts the returned features by 'distance' or 'category'. 
        For request='pois' only.
    :type sortby: string
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean
    
    :rtype: call to Client.request()
    """

    validator.validator(locals(), 'pois')
    
    params = {'request': request,
              'filters': dict(),
              'geometry': dict(),
              }
    
    if request != 'category_list':
        if geojson:
            params['geometry']['geojson'] = geojson
            
        if bbox:
            params['geometry']['bbox'] = bbox
            
        if buffer:
            params['geometry']['buffer'] = buffer
                    
        if filter_category_ids and convert._is_list(filter_category_ids):
            params['filters']['category_ids'] = filter_category_ids
                    
        if filter_category_group_ids and convert._is_list(filter_category_group_ids):
            params['filters']['category_group_ids'] = filter_category_group_ids
                    
        if filters_custom:
            for f in filters_custom:
                params['filters'][f] = filters_custom[f]
                    
        if limit:
            params['limit'] = limit
                    
        if sortby:
            params['sortby'] = sortby
            
    return client.request('/pois', {}, post_json=params, dry_run=dry_run)



# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the distance matrix module."""
import responses
import test as _test

import openrouteservice

class DistanceMatrixTest(_test.TestCase):

    def setUp(self):
        self.key = 'sample_key'
        self.client = openrouteservice.Client(self.key)
        self.coords_valid = [[9.970093,48.477473],
                            [9.207916,49.153868],
                            [37.573242,55.801281],
                            [115.663757,38.106467]]
            
    @responses.activate
    def test_basic_params(self):
        responses.add(responses.POST,
                      'https://api.openrouteservice.org/matrix',
                      body='{"status":"OK","routes":[]}',
                      status=200,
                      content_type='application/json')

        origins = 'all'
        destinations = 'all'
        
        self.client.distance_matrix(self.coords_valid,
                                     sources=origins,
                                     destinations=destinations)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.openrouteservice.org/matrix?api_key={}&profile=driving-car'.format(self.key),
                            responses.calls[0].request.url)

    @responses.activate
    def test_all_params(self):

        origins = [0,1,2]
        destinations = [1,2,3]
        
        query = {'locations': self.coords_valid,
                 'sources': origins,
                 'destinations': destinations,
                 'profile': 'cycling-regular',
                 'metrics': ['duration', 'distance'],
                 'resolve_locations': 'true',
                 'units': 'mi',
                 'optimized': 'false'
                }
        
        payload = {'locations': [[9.970093, 48.477473], [9.207916, 49.153868], [37.573242, 55.801281], [115.663757, 38.106467]],
                   'sources': [0, 1, 2],
                   'destinations': [1, 2, 3],
                   'profile': 'cycling-regular',
                   'metrics': ['duration', 'distance'], 
                   'resolve_locations': 'true',
                   'units': 'mi',
                   'optimized': 'false'}
        
        responses.add(responses.POST,
                      'https://api.openrouteservice.org/matrix',
                      json=query,
                      status=200,
                      content_type='application/json')
        
        resp = self.client.distance_matrix(**query)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.openrouteservice.org/matrix?api_key={}&profile=cycling-regular'.format(self.key),
                            responses.calls[0].request.url)
        self.assertEqual(resp, payload)



# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""
Core client functionality, common across all API requests.
"""

from datetime import datetime
from datetime import timedelta
import functools
import requests
import random
import time
import collections

from openrouteservice import exceptions, __version__

try: # Python 3
    from urllib.parse import urlencode
except ImportError: # Python 2
    from urllib import urlencode


_USER_AGENT = "ORSClientPython.v{}".format(__version__)
_DEFAULT_BASE_URL = "https://api.openrouteservice.org"

_RETRIABLE_STATUSES = set([503])

class Client(object):
    """Performs requests to the ORS API services."""

    def __init__(self, key=None,
                 base_url=_DEFAULT_BASE_URL, 
                 timeout=None, 
                 retry_timeout=60, 
                 requests_kwargs=None,
                 queries_per_minute=40,
                 retry_over_query_limit=False):
        """
        :param key: ORS API key. Required.
        :type key: string

        :param base_url: The base URL for the request. Defaults to the ORS API
            server. Should not have a trailing slash.
        :type base_url: string

        :param timeout: Combined connect and read timeout for HTTP requests, in
            seconds. Specify "None" for no timeout.
        :type timeout: int

        :param retry_timeout: Timeout across multiple retriable requests, in
            seconds.
        :type retry_timeout: int

        :param queries_per_minute: Number of queries per second permitted.
            If the rate limit is reached, the client will sleep for the
            appropriate amount of time before it runs the current query.
            Note, it won't help to initiate another client. This saves you the 
            trouble of raised exceptions.
        :type queries_per_second: int

        :param requests_kwargs: Extra keyword arguments for the requests
            library, which among other things allow for proxy auth to be
            implemented. See the official requests docs for more info:
            http://docs.python-requests.org/en/latest/api/#main-interface
        :type requests_kwargs: dict
        """

        self.session = requests.Session()
        self.key = key
        self.base_url = base_url
        
        self.timeout = timeout
        self.retry_over_query_limit = retry_over_query_limit
        self.retry_timeout = timedelta(seconds=retry_timeout)
        self.requests_kwargs = requests_kwargs or {}
        self.requests_kwargs.update({
            "headers": {"User-Agent": _USER_AGENT,
                        'Content-type': 'application/json'},
            "timeout": self.timeout
        })

        self.queries_per_minute = queries_per_minute
        self.sent_times = collections.deque("", queries_per_minute)

    def request(self, 
                 url, params, 
                 first_request_time=None, 
                 retry_counter=0,
                 requests_kwargs=None, 
                 post_json=None,
                 dry_run=None):
        """Performs HTTP GET/POST with credentials, returning the body as
        JSON.

        :param url: URL path for the request. Should begin with a slash.
        :type url: string

        :param params: HTTP GET parameters.
        :type params: dict or list of key/value tuples

        :param first_request_time: The time of the first request (None if no
            retries have occurred).
        :type first_request_time: datetime.datetime

        :param retry_counter: The number of this retry, or zero for first attempt.
        :type retry_counter: int

        :param requests_kwargs: Same extra keywords arg for requests as per
            __init__, but provided here to allow overriding internally on a
            per-request basis.
        :type requests_kwargs: dict

        :param post_json: HTTP POST parameters. Only specified by calling method.
        :type post_json: dict

        :param dry_run: If 'true', only prints URL and parameters. 'true' or 'false'.
        :type dry_run: string

        :raises ApiError: when the API returns an error.
        :raises Timeout: if the request timed out.
        :raises TransportError: when something went wrong while trying to
            execute a request.
            
        :rtype: dict from JSON response.
        """

        if not first_request_time:
            first_request_time = datetime.now()

        elapsed = datetime.now() - first_request_time
        if elapsed > self.retry_timeout:
            raise exceptions.Timeout()

        if retry_counter > 0:
            # 0.5 * (1.5 ^ i) is an increased sleep time of 1.5x per iteration,
            # starting at 0.5s when retry_counter=1. The first retry will occur
            # at 1, so subtract that first.
            delay_seconds = 1.5 ** (retry_counter - 1)

            # Jitter this value by 50% and pause.
            time.sleep(delay_seconds * (random.random() + 0.5))

        authed_url = self._generate_auth_url(url,
                                             params,
                                             )

        # Default to the client-level self.requests_kwargs, with method-level
        # requests_kwargs arg overriding.
        requests_kwargs = requests_kwargs or {}
        final_requests_kwargs = dict(self.requests_kwargs, **requests_kwargs)

        # Check if the time of the nth previous query (where n is
        # queries_per_second) is under a second ago - if so, sleep for
        # the difference.
        if self.sent_times and len(self.sent_times) == self.queries_per_minute:
            elapsed_since_earliest = time.time() - self.sent_times[0]
            if elapsed_since_earliest < 60:
                print("Request limit of {} per minute exceeded. Wait for {} seconds".format(self.queries_per_minute, 
                                                                                              60 - elapsed_since_earliest))
                time.sleep(60 - elapsed_since_earliest)
        
        # Determine GET/POST.
        # post_json is so far only sent from matrix call
        requests_method = self.session.get
        
        if post_json is not None:
            requests_method = self.session.post
            final_requests_kwargs["json"] = post_json
        
        # Only print URL and parameters for dry_run
        if dry_run:
            print("url:\n{}\nParameters:\n{}".format(self.base_url+authed_url,
                                                    final_requests_kwargs))
            return
        
        try:
            response = requests_method(self.base_url + authed_url,
                                       **final_requests_kwargs)
        except requests.exceptions.Timeout:
            raise exceptions.Timeout()
        except Exception as e:
            raise exceptions.TransportError(e)

        if response.status_code in _RETRIABLE_STATUSES:
            # Retry request.
            print('Server down.\nRetrying for the {}th time.'.format(retry_counter + 1))
            
            return self.request(url, params, first_request_time,
                                 retry_counter + 1, requests_kwargs, post_json)

        try:
            result = self._get_body(response)
            self.sent_times.append(time.time())
            return result
        except exceptions._RetriableRequest as e:
            if isinstance(e, exceptions._OverQueryLimit) and not self.retry_over_query_limit:
                raise
            
            print('Rate limit exceeded.\nRetrying for the {}th time.'.format(retry_counter + 1))
            # Retry request.
            return self.request(url, params, first_request_time,
                                 retry_counter + 1, requests_kwargs,
                                 post_json)
        except:
            raise


    def _get_body(self, response):        
        body = response.json()
#        error = body.get('error')
        status_code = response.status_code
        
        if status_code == 429:
            raise exceptions._OverQueryLimit(
                str(status_code), body)
        if status_code != 200:
            raise exceptions.ApiError(status_code,
                                                       body)

        return body


    def _generate_auth_url(self, path, params):
        """Returns the path and query string portion of the request URL, first
        adding any necessary parameters.

        :param path: The path portion of the URL.
        :type path: string

        :param params: URL parameters.
        :type params: dict or list of key/value tuples

        :rtype: string

        """
        
        if type(params) is dict:
            params = sorted(dict(**params).items())
        
        # Only auto-add API key when using ORS. If own instance, API key must
        # be explicitly added to params
        if self.key:
            params.append(("api_key", self.key))
            return path + "?" + _urlencode_params(params)
        elif self.base_url != _DEFAULT_BASE_URL:
            return path + "?" + _urlencode_params(params)

        raise ValueError("No API key specified. "
                         "Visit https://go.openrouteservice.org/dev-dashboard/ "
                         "to create one.")


from openrouteservice.directions import directions
from openrouteservice.distance_matrix import distance_matrix
from openrouteservice.elevation import elevation_point
from openrouteservice.elevation import elevation_line
from openrouteservice.isochrones import isochrones
from openrouteservice.geocode import pelias_search
from openrouteservice.geocode import pelias_autocomplete
from openrouteservice.geocode import pelias_structured
from openrouteservice.geocode import pelias_reverse
from openrouteservice.places import places


def _make_api_method(func):
    """
    Provides a single entry point for modifying all API methods.
    For now this is limited to allowing the client object to be modified
    with an `extra_params` keyword arg to each method, that is then used
    as the params for each web service request.

    Please note that this is an unsupported feature for advanced use only.
    It's also currently incompatibile with multiple threads, see GH #160.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        args[0]._extra_params = kwargs.pop("extra_params", None)
        result = func(*args, **kwargs)
        try:
            del args[0]._extra_params
        except AttributeError:
            pass
        return result
    return wrapper


Client.directions = _make_api_method(directions)
Client.distance_matrix = _make_api_method(distance_matrix)
Client.elevation_point = _make_api_method(elevation_point)
Client.elevation_line = _make_api_method(elevation_line)
Client.isochrones = _make_api_method(isochrones)
Client.pelias_search = _make_api_method(pelias_search)
Client.pelias_autocomplete = _make_api_method(pelias_autocomplete)
Client.pelias_structured = _make_api_method(pelias_structured)
Client.pelias_reverse = _make_api_method(pelias_reverse)
Client.places = _make_api_method(places)


def _urlencode_params(params):
    """URL encodes the parameters.

    :param params: The parameters
    :type params: list of key/value tuples.

    :rtype: string
    """
    # urlencode does not handle unicode strings in Python 2.
    # Firstly, normalize the values so they get encoded correctly.
    params = [(key, _normalize_for_urlencode(val)) for key, val in params]
    # Secondly, unquote unreserved chars which are incorrectly quoted
    # by urllib.urlencode, causing invalid auth signatures. See GH #72
    # for more info.
    return requests.utils.unquote_unreserved(urlencode(params))


try:
    unicode
    # NOTE(cbro): `unicode` was removed in Python 3. In Python 3, NameError is
    # raised here, and caught below.

    def _normalize_for_urlencode(value):
        """(Python 2) Converts the value to a `str` (raw bytes)."""
        if isinstance(value, unicode):
            return value.encode('utf8')

        if isinstance(value, str):
            return value

        return _normalize_for_urlencode(str(value))

except NameError:
    def _normalize_for_urlencode(value):
        """(Python 3) No-op."""
        # urlencode in Python 3 handles all the types we are passing it.
        return value



# -*- coding: utf-8 -*-
# Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the distance matrix module."""
import responses
import test as _test

import openrouteservice


class DistanceMatrixTest(_test.TestCase):

    def setUp(self):
        self.key = 'sample_key'
        self.client = openrouteservice.Client(self.key)
        self.coords_valid = [[9.970093, 48.477473],
                             [9.207916, 49.153868],
                             [37.573242, 55.801281],
                             [115.663757, 38.106467],
                             [8.34234, 48.23424]]

    @responses.activate
    def test_basic_params(self):
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/isochrones',
                      body='{"status":"OK","routes":[]}',
                      status=200,
                      content_type='application/json')

        isochrone = self.client.isochrones(self.coords_valid[0])

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.openrouteservice.org/isochrones?api_key={}&'
                            'locations=9.970093%2C48.477473&'
                            'profile=driving-car&range=60&interval=30'.format(self.key),
                            responses.calls[0].request.url)

    @responses.activate
    def test_all_params(self):
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/isochrones',
                      body='{"status":"OK","routes":[]}',
                      status=200,
                      content_type='application/json')

        isochrone = self.client.isochrones(self.coords_valid,
                                           profile='cycling-regular',
                                           range_type='distance',
                                           intervals=[1000, 2000],
                                           units='m',
                                           location_type='destination',
                                           smoothing=0.5,
                                           attributes=['area', 'reachfactor']
                                           )

        iso_parameter = {'locations': [[9.970093, 48.477473], [9.207916, 49.153868]],
                         'profile': 'cycling-regular',
                         'range_type': 'distance',
                         'range': [1000, 2000],
                         'units': 'm',
                         'location_type': 'destination',
                         'attributes': ['area', 'reachfactor'],
                         'interval': [30]
                         }

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.openrouteservice.org/isochrones?api_key={}&'
                            'locations=9.970093%2C48.477473%7C9.207916'
                            '%2C49.153868%7C37.573242%2C55.801281%7C115.663757'
                            '%2C38.106467%7C8.34234%2C48.23424&profile=cycling-regular&'
                            'range_type=distance&range=1000%2C2000&'
                            'units=m&location_type=destination&'
                            'smoothing=0.5&'
                            'attributes=area|reachfactor&interval=30'.format(self.key),
                            responses.calls[0].request.url)



# -*- coding: utf-8 -*-
# Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the validator module."""

from openrouteservice import validator, exceptions
import test as _test
from test.test_helper import *

class ValidatorTest(_test.TestCase):
    
    def test_directions_correct_schema(self):
        validator.validator(ENDPOINT_DICT['directions'], 'directions')

    def test_directions_wrong_schema(self):
        ENDPOINT_DICT['directions']['preference'] = 'best'
        ENDPOINT_DICT['directions']['attributeds'] = ['avgspeed']
        ENDPOINT_DICT['directions']['radiuses'] = 2 * PARAM_LIST_ONE
        ENDPOINT_DICT['directions']['options'] = {'maximum_speed': '20'}

        with self.assertRaises(exceptions.ValidationError) as e:
            validator.validator(ENDPOINT_DICT['directions'], 'directions')
        em = e.exception
        print(em)
        
        self.assertIn('unallowed value best', str(em))
        self.assertIn('unknown field', str(em))
        self.assertIn('max length is 2', str(em))
        self.assertIn('must be of integer type', str(em))
        
    def test_isochrones_correct_schema(self):
        validator.validator(ENDPOINT_DICT['isochrones'], 'isochrones')

    def test_isochrones_wrong_schema(self):
        del ENDPOINT_DICT['isochrones']['locations']
        ENDPOINT_DICT['isochrones'].update(
                { 'attributes': ['areas', 'reachfactor'],
                  'intervals': ['30']
                  })
        
        with self.assertRaises(exceptions.ValidationError) as e:
            validator.validator(ENDPOINT_DICT['isochrones'], 'isochrones')
        em = e.exception        
        print(em)
        
        self.assertIn('required field', str(em))
        self.assertIn('unallowed value', str(em))
        self.assertIn('must be of integer type', str(em))

    def test_distance_matrix_correct_schema(self):
        validator.validator(ENDPOINT_DICT['distance_matrix'], 'distance_matrix')

    def test_distance_matrix_wrong_schema(self):
        ENDPOINT_DICT['distance_matrix'].update(
                { 'sources': [0, 1, 2],
                  'metrics': ['duration', 'distance'],
                  'resolve_locations': 'true',
                  'units': 'm',
                  'optimizedd': 'false'}
                )
        
        with self.assertRaises(exceptions.ValidationError) as e:
            validator.validator(ENDPOINT_DICT['distance_matrix'], 'distance_matrix')
        em = e.exception        
        print(em)
        
        self.assertIn('no definitions validate', str(em))
        self.assertIn('unknown field', str(em))

    def test_search_correct_schema(self):
        validator.validator(ENDPOINT_DICT['pelias_search'], 'pelias_search')

    def test_search_wrong_schema(self):
        ENDPOINT_DICT['pelias_search'].update({
                  'layers': ['locality', 'name'],
                  'circle_radius': {50},
                })
        
        with self.assertRaises(exceptions.ValidationError) as e:
            validator.validator(ENDPOINT_DICT['pelias_search'], 'pelias_search')
        em = e.exception        
        print(em)
        
        self.assertIn("unallowed values ['name']", str(em))
        self.assertIn('must be of integer type', str(em))

    def test_autocomplete_correct_schema(self):
        validator.validator(ENDPOINT_DICT['pelias_autocomplete'], 'pelias_autocomplete')

    def test_autocomplete_wrong_schema(self):
        ENDPOINT_DICT['pelias_autocomplete'].update({
                  'layers': ['locality', 'name'],
                })
        
        with self.assertRaises(exceptions.ValidationError) as e:
            validator.validator(ENDPOINT_DICT['pelias_autocomplete'], 'pelias_autocomplete')
        em = e.exception        
        print(em)
        
        self.assertIn("unallowed values ['name']", str(em))
        
    def test_structured_correct_schema(self):
        validator.validator(ENDPOINT_DICT['pelias_structured'], 'pelias_structured')
        
    def test_structured_wrong_schema(self):
        ENDPOINT_DICT['pelias_structured'].update({
                'address': {'Berliner Straße 45'},
                'postalcode': '69120',                
                })
        params = {'address': {'Berliner Straße 45'},
                  'neighbourhood': 'Neuenheimer Feld',
                  'borough': 'Heidelberg',
                  'locality': 'Heidelberg',
                  'county': 'Rhein-Neckar-Kreis',
                  'region': 'Baden-Württemberg',
                  'postalcode': '69120',
                  'country': 'de',
                  }
        
        with self.assertRaises(exceptions.ValidationError) as e:
            validator.validator(ENDPOINT_DICT['pelias_structured'], 'pelias_structured')
        em = e.exception        
        print(em)
        
        self.assertIn('must be of integer type', str(em))
        self.assertIn('must be of string type', str(em))

    def test_reverse_correct_schema(self):
        validator.validator(ENDPOINT_DICT['pelias_reverse'], 'pelias_reverse')
        
    def test_reverse_wrong_schema(self):
        ENDPOINT_DICT['pelias_reverse'].update({
                  'country': 35,
                  'sources': ['osm', 'wof', 'gm'],                
                })
        
        with self.assertRaises(exceptions.ValidationError) as e:
            validator.validator(ENDPOINT_DICT['pelias_reverse'], 'pelias_reverse')
        em = e.exception        
        print(em)
        
        self.assertIn('must be of string type', str(em))
        self.assertIn("unallowed values ['gm']", str(em))
        
    def test_pois_correct_schema(self):
        validator.validator(ENDPOINT_DICT['pois'], 'pois')
        
        ENDPOINT_DICT['pois']['geojson'] = PARAM_GEOJSON_LINE
        validator.validator(ENDPOINT_DICT['pois'], 'pois')
        
        ENDPOINT_DICT['pois']['geojson'] = PARAM_GEOJSON_POLY
        validator.validator(ENDPOINT_DICT['pois'], 'pois')
        
        ENDPOINT_DICT['pois']['request'] = 'stats'
        validator.validator(ENDPOINT_DICT['pois'], 'pois')
    
        ENDPOINT_DICT['pois'] = {'request': 'list'}
        validator.validator(ENDPOINT_DICT['pois'], 'pois')        
        
    def test_pois_wrong_schema(self):
        ENDPOINT_DICT['pois'].update({
                  'geojson': {'type': 'Point'},
                  'filters_custom': {'wheelchair': ['maybe']},
                  'limit': 1200,
                })
        
        with self.assertRaises(exceptions.ValidationError) as e:
            validator.validator(ENDPOINT_DICT['pois'], 'pois')
        em = e.exception        
        print(em)
        
        self.assertIn("field 'coordinates' is required", str(em))
        self.assertIn('max value is 1000', str(em))
        self.assertIn('unallowed value maybe', str(em))



# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the ORS Matrix API."""

from openrouteservice import convert, validator


def distance_matrix(client, locations,
                    profile='driving-car',
                    sources='all',
                    destinations='all',
                    metrics=None,
                    resolve_locations=None,
                    units=None,
                    optimized=True,
                    dry_run=None):
    """ Gets travel distance and time for a matrix of origins and destinations.

    :param locations: One or more pairs of lng/lat values.
    :type locations: a single location, or a list of locations, where a
        location is a list or tuple of lng,lat values

    :param profile: Specifies the mode of transport to use when calculating
        directions. One of ["driving-car", "driving-hgv", "foot-walking",
        "foot-hiking", "cycling-regular", "cycling-road",
        "cycling-safe", "cycling-mountain", "cycling-tour", 
        "cycling-electric",]. Default "driving-car".
    :type profile: string

    :param sources: A list of indices that refer to the list of locations
        (starting with 0) or 'all'. Default 'all'.
    :type sources: One or more indices inside a list; or 'all' (string).

    :param destinations: A list of indices that refer to the list of locations
        (starting with 0) or 'all'. Default 'all'.
    :type destinations: One or more indices inside a list; or 'all' (string).

    :param metrics: Specifies a list of returned metrics. One or more of ["distance",
        "duration"]. Default ['duration'].
    :type metrics: list of strings

    :param resolve_locations: Specifies whether given locations are resolved or
        not. If set 'true', every element in destinations and sources will 
        contain the name element that identifies the name of the closest street.
        Default False.
    :type resolve_locations: boolean

    :param units: Specifies the unit system to use when displaying results.
        One of ["m", "km", "m"]. Default "m".
    :type units: string

    :param optimized: Specifies whether Dijkstra algorithm ('false') or any 
        available technique to speed up shortest-path routing ('true') is used. 
        For normal Dijkstra the number of visited nodes is limited to 100000.
        Default True
    :type optimized: boolean
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean
    
    :raises ValueError: When profile parameter has wrong value.
    
    :rtype: call to Client.request()
    """

    validator.validator(locals(), 'distance_matrix')
    
    params = {
            "locations": locations,
            "sources": sources,
            "destinations": destinations
            }
    
    get_params = {}

    if profile:
        params["profile"] = profile
        get_params['profile'] = profile

    if sources:
        if sources == 'all': 
            params["sources"] = sources
        else:
            params["sources"] = convert._comma_list(sources)

    if destinations:
        if destinations == 'all': 
            params["destinations"] = destinations
        else:
            params["destinations"] = convert._comma_list(destinations)

    if metrics:
        params["metrics"] = convert._pipe_list(metrics)

    if resolve_locations:
        params["resolve_locations"] = resolve_locations

    if units:
        params["units"] = units

    if optimized:
        # not checked on backend, check here
        # convert._checkBool(optimized)
        params["optimized"] = optimized


    return client.request("/matrix", get_params, post_json=params, dry_run=dry_run) 


# -*- coding: utf-8 -*-

# Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Performs requests to the ORS elevation API."""

from openrouteservice import validator

def elevation_point(client, format_in, geometry,
                    format_out='geojson',
                    dataset='srtm',
                    dry_run=None):
    """
    POSTs 2D point to be enriched with elevation.
    
    :param format_in: Format of input geometry. One of ['geojson',
        'point']
    :type format_in: string
    
    :param geometry: Point geometry 
    :type geometry: depending on format_in, either list of coordinates or Point geojson
    
    :param format_out: Format of output geometry, one of ['geojson', 'point']
    :type format_out: string
    
    :param dataset: Elevation dataset to be used. Currently only SRTM v4.1 available.
    :type dataset: string
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean
    
    :returns: correctly formatted parameters
    :rtype: Client.request()
    """
    
    validator.validator(locals(), 'elevation_point')

    params = {'format_in': format_in,
              'geometry': geometry,
              'format_out': format_out,
              'dataset': dataset}    
            
    return client.request('/elevation/point', {}, post_json=params, dry_run=dry_run)

def elevation_line(client, format_in, geometry,
                    format_out='geojson',
                    dataset='srtm',
                    dry_run=None):
    """
    POSTs 2D point to be enriched with elevation.
    
    :param format_in: Format of input geometry. One of ['geojson',
        'polyline', 'encodedpolyline']
    :type format_in: string
    
    :param geometry: Point geometry 
    :type geometry: depending on format_in, either list of coordinates, LineString
        geojson or string
    
    :param format_out: Format of output geometry, one of ['geojson',
        'polyline', 'encodedpolyline']
    :type format_out: string
    
    :param dataset: Elevation dataset to be used. Currently only SRTM v4.1 available.
    :type dataset: string
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean
    
    :returns: correctly formatted parameters
    :rtype: Client.request()
    """
    
    validator.validator(locals(), 'elevation_line')

    params = {'format_in': format_in,
              'geometry': geometry,
              'format_out': format_out,
              'dataset': dataset}    
            
    return client.request('/elevation/point', {}, post_json=params, dry_run=dry_run)


# -*- coding: utf-8 -*-

# Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Tests for the distance matrix module."""
import responses
import test as _test

import openrouteservice
from test.test_helper import ENDPOINT_DICT

class ElevationTest(_test.TestCase):
    def setUp(self):
        self.key = 'sample_key'
        self.client = openrouteservice.Client(self.key)
        self.coords = coords = [[13.331302, 38.108433],
                                [13.331273, 38.108493]]
        
#    @responses.activate
#    def test_line(self):
#        query = {'format_in': 'geojson',
#                 'format_out': 'polyline',
#                 'geometry': {'type': 'LineString', 'coordinates': self.coords},
#                 'dataset': 'srtm'
#                }
#        
#        responses.add(responses.POST,
#                      'https://api.openrouteservice.org/elevation/line',
#                      json=query,
#                      status=200,
#                      content_type='application/json')
#        
#        resp = self.client.elevation_point(**query)
#        
#        self.assertEquals(len(responses.calls), 1)
#        self.assertURLEqual('https://api.openrouteservice.org/elevation/line?api_key={}'.format(self.key),
#                            responses.calls[0].request.url)        
#        self.assertEquals(responses.calls[0].response.json(), query)
        
    @responses.activate
    def test_point(self):        
        query = ENDPOINT_DICT['elevation_point']
        responses.add(responses.POST,
                      'https://api.openrouteservice.org/elevation/point',
                      json= query,
                      status=200,
                      content_type='application/json')
        
        resp = self.client.elevation_point(**query)
        
        self.assertEquals(len(responses.calls), 1)
        self.assertURLEqual('https://api.openrouteservice.org/elevation/point?api_key={}'.format(self.key),
                            responses.calls[0].request.url)        
        self.assertEquals(responses.calls[0].response.json(), query)
    


# -*- coding: utf-8 -*-
# Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the ORS isochrones API."""

from openrouteservice import convert, validator


def isochrones(client, locations,
               profile='driving-car',
               range_type=None,
               intervals=[60],
               segments=30,
               units=None,
               location_type=None,
               smoothing=None,
               attributes=None,
               # options=None,
               intersections=None,
               dry_run=None):
    """ Gets travel distance and time for a matrix of origins and destinations.

    :param locations: One pair of lng/lat values.
    :type locations: list or tuple of lng,lat values

    :param profile: Specifies the mode of transport to use when calculating
        directions. One of ["driving-car", "driving-hgv", "foot-walking",
        "foot-hiking", "cycling-regular", "cycling-road",
        "cycling-safe", "cycling-mountain", "cycling-tour", 
        "cycling-electric",]. Default "driving-car".
    :type profile: string

    :param range_type: Set 'time' for isochrones or 'distance' for equidistants.
        Default 'time'.
    :type sources: string

    :param intervals: Ranges to calculate distances/durations for. This can be
        a list of multiple ranges, e.g. [600, 1200, 1400] or a single value list.
        In the latter case, you can also specify the 'segments' variable to break
        the single value into more isochrones. In meters or seconds. Default [60].
    :type intervals: list of integer(s)

    :param segments: Segments isochrones or equidistants for one 'intervals' value.
        Only has effect if used with a single value 'intervals' parameter.
        In meters or seconds. Default 20.
    :type segments: integer

    :param units: Specifies the unit system to use when displaying results.
        One of ["m", "km", "m"]. Default "m".
    :type units: string
    
    :param location_type: 'start' treats the location(s) as starting point,
        'destination' as goal. Default 'start'.
    :type location_type: string

    :param smoothing: Applies a level of generalisation to the isochrone polygons generated.
        Value between 0 and 1, whereas a value closer to 1 will result in a more generalised shape.
    :type smoothing: float

    :param attributes: 'area' returns the area of each polygon in its feature
        properties. 'reachfactor' returns a reachability score between 0 and 1.
        'total_pop' returns population statistics from https://ghsl.jrc.ec.europa.eu/about.php.
        One or more of ['area', 'reachfactor', 'total_pop']. Default 'area'.
    :type attributes: list of string(s)

    :param options: not implemented right now.
    :type options: dict
    
    :param intersections: not implented right now.
    :type intersections: boolean as string
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean
    
    :raises ValueError: When parameter has invalid value(s).
    
    :rtype: call to Client.request()
    """

    validator.validator(locals(), 'isochrones')

    params = {
        "locations": convert._build_coords(locations)
    }

    if profile:
        params["profile"] = profile

    if range_type:
        params["range_type"] = range_type

    if intervals:
        params["range"] = convert._comma_list(intervals)

    if segments:
        params["interval"] = str(segments)
        
    if units:
        # if units and (range_type == None or range_type == 'time'):
        #     raise ValueError("For range_type time, units cannot be specified.")
        params["units"] = units

    if location_type:
        params["location_type"] = location_type

    if smoothing:
        params["smoothing"] = convert._format_float(smoothing)

    if attributes:
        params["attributes"] = convert._pipe_list(attributes)

    return client.request("/isochrones", params, dry_run=dry_run)


# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

__version__ = "1.1.2"

# Make sure QGIS plugin can import openrouteservice-py

from openrouteservice.client import Client
## Allow sphinx to pick up these symbols for the documentation.
#__all__ = ["Client"]



# -*- coding: utf-8 -*-
# Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Performs requests to the ORS geocode API (direct Pelias clone)."""
from openrouteservice import convert, validator


def pelias_search(client, text,
                  focus_point=None,
                  rect_min_x=None,
                  rect_min_y=None,
                  rect_max_x=None,
                  rect_max_y=None,
                  circle_point=None,
                  circle_radius=None,
                  sources=None,
                  layers=None,
                  country=None,
                  size=None,
                  dry_run=None):
    """
    Geocoding is the process of converting addresses into geographic
    coordinates.

    This endpoint queries directly against a Pelias instance.

    :param text: Full-text query against search endpoint. Required.
    :type text: string

    :param focus_point: Focusses the search to be around this point and gives
        results within a 100 km radius higher scores.
    :type query: list or tuple of (Long, Lat)

    :param rect_min_x: Min longitude by which to constrain request geographically.
    :type rect_min_x: float

    :param rect_min_y: Min latitude by which to constrain request geographically.
    :type rect_min_y: float

    :param rect_max_x: Max longitude by which to constrain request geographically.
    :type rect_max_x: float

    :param rect_max_y: Max latitude by which to constrain request geographically.
    :type rect_max_y: float

    :param circle_point: Geographical constraint in form a circle.
    :type circle_point: list or tuple of (Long, Lat)

    :param circle_radius: Radius of circle constraint in km. Default 50.
    :type circle_radius: integer

    :param sources: The originating source of the data. One or more of
        ['osm', 'oa', 'wof', 'gn']. Currently only 'osm', 'wof' and 'gn' are
        supported.
    :type sources: list of strings

    :param layers: The administrative hierarchy level for the query. Refer to
        https://github.com/pelias/documentation/blob/master/search.md#filter-by-data-type
        for details.
    :type layers: list of strings

    :param country: Constrain query by country. Accepts alpha-2 or alpha-3
        digit ISO-3166 country codes.
    :type country: list of strings

    :param size: The amount of results returned. Default 10.
    :type size: integer
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean

    :raises ValueError: When parameter has invalid value(s).
    :raises TypeError: When parameter is of the wrong type.

    :rtype: call to Client.request()
    """

    validator.validator(locals(), 'pelias_search')

    params = {'text': text}

    if focus_point:
        params['focus.point.lon'] = convert._format_float(focus_point[0])
        params['focus.point.lat'] = convert._format_float(focus_point[1])

    if rect_min_x:
        params['boundary.rect.min_lon'] = convert._format_float(rect_min_x)  #

    if rect_min_y:
        params['boundary.rect.min_lat'] = convert._format_float(rect_min_y)  #

    if rect_max_x:
        params['boundary.rect.max_lon'] = convert._format_float(rect_max_x)  #

    if rect_max_y:
        params['boundary.rect.max_lat'] = convert._format_float(rect_max_y)  #

    if circle_point:
        params['boundary.circle.lon'] = convert._format_float(circle_point[0])  #
        params['boundary.circle.lat'] = convert._format_float(circle_point[1])  #

    if circle_radius:
        params['boundary.circle.radius'] = circle_radius

    if sources:
        params['sources'] = convert._comma_list(sources)

    if layers:
        params['layers'] = convert._comma_list(layers)

    if country:
        params['country'] = country

    if size:
        params['size'] = size

    return client.request("/geocode/search", params, dry_run=dry_run)


def pelias_autocomplete(client, text,
                        focus_point=None,
                        rect_min_x=None,
                        rect_min_y=None,
                        rect_max_x=None,
                        rect_max_y=None,
                        country=None,
                        sources=None,
                        layers=None,
                        dry_run=None):
    """
    Autocomplete geocoding can be used alongside /search to enable real-time feedback.
    It represents a type-ahead functionality, which helps to find the desired location,
    without to require a fully specified search term.

    This endpoint queries directly against a Pelias instance.
    For fully documentation, please see https://github.com/pelias/documentation/blob/master/autocomplete.md

    :param text: Full-text query against search endpoint. Required.
    :type text: string

    :param focus_point: Focusses the search to be around this point and gives
        results within a 100 km radius higher scores.
    :type query: list or tuple of (Long, Lat)

    :param rect_min_x: Min longitude by which to constrain request geographically.
    :type rect_min_x: float

    :param rect_min_y: Min latitude by which to constrain request geographically.
    :type rect_min_y: float

    :param rect_max_x: Max longitude by which to constrain request geographically.
    :type rect_max_x: float

    :param rect_max_y: Max latitude by which to constrain request geographically.
    :type rect_max_y: float

    :param country: Constrain query by country. Accepts alpha-2 or alpha-3
        digit ISO-3166 country codes.
    :type country: list of strings

    :param sources: The originating source of the data. One or more of
        ['osm', 'oa', 'wof', 'gn']. Currently only 'osm', 'wof' and 'gn' are
        supported.
    :type sources: list of strings

    :param layers: The administrative hierarchy level for the query. Refer to
        https://github.com/pelias/documentation/blob/master/search.md#filter-by-data-type
        for details.
    :type layers: list of strings
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean

    :raises ValueError: When parameter has invalid value(s).
    :raises TypeError: When parameter is of the wrong type.

    :rtype: dict from JSON response
    """

    validator.validator(locals(), 'pelias_autocomplete')

    params = {'text': text}

    if focus_point:
        params['focus.point.lon'] = convert._format_float(focus_point[0])
        params['focus.point.lat'] = convert._format_float(focus_point[1])

    if rect_min_x:
        params['boundary.rect.min_lon	'] = convert._format_float(rect_min_x)

    if rect_min_y:
        params['boundary.rect.min_lat	'] = convert._format_float(rect_min_y)

    if rect_max_x:
        params['boundary.rect.max_lon	'] = convert._format_float(rect_max_x)

    if rect_max_y:
        params['boundary.rect.max_lon	'] = convert._format_float(rect_max_y)

    if country:
        params['boundary.country'] = country

    if sources:
        params['sources'] = convert._comma_list(sources)

    if layers:
        params['layers'] = convert._comma_list(layers)

    return client.request("/geocode/autocomplete", params, dry_run=dry_run)


def pelias_structured(client,
                      address=None,
                      neighbourhood=None,
                      borough=None,
                      locality=None,
                      county=None,
                      region=None,
                      postalcode=None,
                      country=None,
                      dry_run=None,
                      # size=None
                      ):
    """
    With structured geocoding, you can search for the individual parts of a location.
    Structured geocoding is an option on the search endpoint,
    which allows you to define a query that maintains the individual fields.

    This endpoint queries directly against a Pelias instance.
    For full documentation, please see https://github.com/pelias/documentation/blob/master/structured-geocoding.md

    :param address: Can contain a full address with house number or only a street name.
    :type address: string

    :param neighbourhood: Neighbourhoods are vernacular geographic entities that
        may not necessarily be official administrative divisions but are important nonetheless.
    :type neighbourhood: string

    # Check all passed arguments
    convert._is_valid_args(locals())

    :param borough: Mostly known in the context of New York City, even though they may exist in other cities.
    :type borough: string

    :param locality: Localities are equivalent to what are commonly referred to as cities.
    :type locality: string

    :param county: Administrative divisions between localities and regions.
        Not as commonly used in geocoding as localities, but useful when attempting to
        disambiguate between localities.
    :type county: string

    :param region: Normally the first-level administrative divisions within countries, analogous to states
        and provinces in the United States and Canada. Can be a full name or abbreviation.
    :type region: string

    :param postalcode: Dictated by an administrative division, which is almost always countries.
        Postal codes are unique within a country.
    :type postalcode: integer

    :param country: Highest-level divisions supported in a search. Can be a full name or abbreviation.
    :type country: string
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean

    :raises TypeError: When parameter is of the wrong type.

    :rtype: dict from JSON response
    """

    validator.validator(locals(), 'pelias_structured')

    params = dict()

    if address:
        params['address'] = address

    if neighbourhood:
        params['neighbourhood'] = neighbourhood

    if borough:
        params['borough'] = borough

    if locality:
        params['locality'] = locality

    if county:
        params['county'] = county

    if region:
        params['region'] = region

    if postalcode:
        params['postalcode'] = postalcode

    if country:
        params['country'] = country

    return client.request("/geocode/search/structured", params, dry_run=dry_run)


def pelias_reverse(client, point,
                   circle_radius=None,
                   sources=None,
                   layers=None,
                   country=None,
                   size=None,
                   dry_run=None):
    """
    Reverse geocoding is the process of converting geographic coordinates into a
    human-readable address.

    This endpoint queries directly against a Pelias instance.

    :param point: Coordinate tuple. Required.
    :type point: list or tuple of [Lon, Lat]

    :param circle_radius: Radius around point to limit query in km. Default 1.
    :type circle_radius: integer

    :param sources: The originating source of the data. One or more of
        ['osm', 'oa', 'wof', 'gn']. Currently only 'osm', 'wof' and 'gn' are
        supported.
    :type sources: list of strings

    :param layers: The administrative hierarchy level for the query. Refer to
        https://github.com/pelias/documentation/blob/master/search.md#filter-by-data-type
        for details.
    :type layers: list of strings

    :param country: Constrain query by country. Accepts alpha-2 or alpha-3
        digit ISO-3166 country codes.
    :type country: list of strings

    :param size: The amount of results returned. Default 10.
    :type size: integer
    
    :param dry_run: Print URL and parameters without sending the request.
    :param dry_run: boolean

    :raises ValueError: When parameter has invalid value(s).

    :rtype: dict from JSON response
    """

    validator.validator(locals(), 'pelias_reverse')

    params = dict()

    params['point.lon'] = convert._format_float(point[0])
    params['point.lat'] = convert._format_float(point[1])

    if circle_radius:
        params['boundary.circle.radius'] = str(circle_radius)

    if sources:
        params['sources'] = convert._comma_list(sources)

    if layers:
        params['layers'] = convert._comma_list(layers)

    if country:
        params['country'] = country

    if size:
        params['size'] = size

    return client.request("/geocode/reverse", params, dry_run=dry_run)



# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""
Defines exceptions that are thrown by the ORS client.
"""

class ValidationError(Exception):
    """Something went wrong during cerberus validation"""
    def __init__(self, errors):
        for error in errors:
            msg = '\n'.join(["Argument '{}': {}".format(error, errors[error][0]) for error in errors])
        Exception.__init__(self, msg)

class ApiError(Exception):
    """Represents an exception returned by the remote API."""
    def __init__(self, status, message=None):
        self.status = status
        self.message = message

    def __str__(self):
        if self.message is None:
            return self.status
        else:
            return "%s (%s)" % (self.status, self.message)

class TransportError(Exception):
    """Something went wrong while trying to execute the request."""

    def __init__(self, base_exception=None):
        self.base_exception = base_exception

    def __str__(self):
        if self.base_exception:
            return str(self.base_exception)

        return "An unknown error occurred."

class HTTPError(Exception):
    """An unexpected HTTP error occurred."""
    def __init__(self, status_code):
        self.status_code = status_code

    def __str__(self):
        return "HTTP Error: %d" % self.status_code

class Timeout(Exception):
    """The request timed out."""
    pass

class _RetriableRequest(Exception):
    """Signifies that the request can be retried."""
    pass

class _OverQueryLimit(ApiError, _RetriableRequest):
    """Signifies that the request failed because the client exceeded its query rate limit.

    Normally we treat this as a retriable condition, but we allow the calling code to specify that these requests should
    not be retried.
    """
    pass



# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#


"""Tests for client module."""

import responses
import time

import openrouteservice
import test as _test

class ClientTest(_test.TestCase):
    
    def setUp(self):
        self.key = 'sample_key'
        self.query = 'Heidelberg'
        self.coords_valid = ((8.34234,48.23424),(8.34423,48.26424))

    def test_no_api_key(self):
        with self.assertRaises(Exception):
            client = openrouteservice.Client()
            client.directions(self.coords_valid)

    def test_invalid_api_key(self):
        with self.assertRaises(Exception):
            client = openrouteservice.Client(key="Invalid key.")
            client.directions(self.coords_valid)

    def test_urlencode(self):
        encoded_params = openrouteservice.client._urlencode_params([("address", "=Sydney ~")])
        self.assertEqual("address=%3DSydney+~", encoded_params)

    @responses.activate
    def test_queries_per_minute_sleep_function(self):
        # This test assumes that the time to run a mocked query is
        # relatively small, eg a few milliseconds. We define a rate of
        # 3 queries per second, and run double that, which should take at
        # least 1 minute but no more than 2.
        queries_per_minute = 2
        query_range = range(queries_per_minute * 2)
        
        for _ in query_range:
            responses.add(responses.GET,
                          'https://api.openrouteservice.org/directions',
                          body='{"status":"OK","results":[]}',
                          status=200,
                          content_type='application/json')
            
        client = openrouteservice.Client(key="58d904a497c67e00015b45fcc8de25513ac6446692280cb066e66e8c",
                                   queries_per_minute=queries_per_minute)
        start = time.time()
        for idx, _ in enumerate(query_range):
            client.directions(self.coords_valid)
        end = time.time()
        self.assertTrue(start + 60 < end < start + 120)
        
    def test_overquerylimit_error(self):
        # Assume more queries_per_minute than allowed by API policy and 
        # don't allow retries if API throws 'rate exceeded' error, which 
        # should be caught.
        queries_per_minute = 110
        query_range = range(queries_per_minute * 2)
            
        client = openrouteservice.Client(key='5b3ce3597851110001cf624870cf2f2a58d44c718542b3088221b684',
                                   queries_per_minute=queries_per_minute,
                                   retry_over_query_limit=False)
        
        with self.assertRaises(openrouteservice.exceptions._OverQueryLimit):
            for _ in query_range:
                client.directions(self.coords_valid)

    @responses.activate
    def test_raise_timeout_retriable_requests(self):
        # Mock query gives 503 as HTTP status, code should try a few times to 
        # request the same and then fail on Timout() error.
        retry_timeout = 3
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/directions',
                      body='{"status":"OK","results":[]}',
                      status=503,
                      content_type='application/json')
            
        client = openrouteservice.Client(key=self.key, 
                                         retry_timeout=retry_timeout)
        
        start = time.time()
        with self.assertRaises(openrouteservice.exceptions.Timeout):
            client.directions(self.coords_valid)
        end = time.time()
        self.assertTrue(retry_timeout < end-start < 2 * retry_timeout)

    @responses.activate
    def test_host_override_with_parameters(self):
        # Test if it's possible to override host for individual hosting.
        responses.add(responses.GET,
                      "https://foo.com/bar",
                      body='{"status":"OK","results":[]}',
                      status=200,
                      content_type="application/json")

        client = openrouteservice.Client(base_url="https://foo.com")
        client.request("/bar", {'bunny':'pretty', 'fox':'prettier'})
        
        self.assertURLEqual("https://foo.com/bar?bunny=pretty&fox=prettier",
                            responses.calls[0].request.url)
        self.assertEqual(1, len(responses.calls))
    
    @responses.activate
    def test_dry_run(self):
        # Test that nothing is requested when dry_run is 'true'
        
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/directions',
                      body='{"status":"OK","results":[]}',
                      status=200,
                      content_type='application/json')
        
        client = openrouteservice.Client(key=self.key)
        req = client.request(params={'format_out': 'geojson'},
                             url='directions/',
                             dry_run='true')
        
        self.assertEqual(0, len(responses.calls))


# -*- coding: utf-8 -*-
import sys


try:
  from setuptools import setup
except ImportError:
  from distutils.core import setup

def readme():
    with open('README.rst') as f:
        return f.read()

if sys.version_info <= (2, 6):
  error = 'Requires Python Version 2.7 or above... exiting.'
  print >> sys.stderr, error
  sys.exit(1)

setup(name='openrouteservice',
      version='1.1.2',
      description='Python client for requests to openrouteservice API services',
      long_description=readme(),
      classifiers=[
              'Development Status :: 4 - Beta',
              'License :: OSI Approved :: Apache Software License',
              'Operating System :: OS Independent',
              'Programming Language :: Python :: 2.7',
              'Programming Language :: Python :: 3.4',
              'Programming Language :: Python :: 3.5',
              'Programming Language :: Python :: 3.6',
              ],
      keywords='routing accessibility router OSM ORS openrouteservice openstreetmap isochrone POI',
      url='https://github.com/GIScience/openrouteservice-py',
      author='Nils Nolde',
      author_email='nils.nolde@gmail.com',
      license='Apache-2.0',
      packages=['openrouteservice'],
      install_requires = [
              'requests==2.19.1',
              'Cerberus==1.2',
              'certifi==2018.8.24',
              'chardet==3.0.4',
              'cookies==2.2.1',
              'idna==2.7',
              'six==1.11.0',
              'urllib3==1.23'],
      include_package_data=True,
      test_suite='nose.collector',
      tests_require=['nose==1.3.7',
                     'responses==0.9.0',
                     'mock==1.23'],
      zip_safe=False)



# -*- coding: utf-8 -*-

# Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

"""Validates the used parameter with Cerberus."""

from cerberus import Validator, TypeDefinition
from openrouteservice import exceptions

# Add the tuple type
tuple_type = TypeDefinition("tuple", (tuple), ())
Validator.types_mapping['tuple'] = tuple_type
v = Validator()


def validator(args, function):
    """
    Validates arguments against function specific schemas
    
    :param args: Arguments to validate
    :type args: dict
    
    :param function: calling function
    :type function: string
    
    :raises: ValidationError    
    """
    
    # Sanitize locals() variables
    args = {arg: args[arg] for arg in args if arg!='client' and args[arg] is not None}
        
    if function == 'directions':
        coords_len = len(args['coordinates'])
        val = _directions_validation(args, coords_len)
    elif function == "isochrones":
        val = _isochrones_validation(args)
    elif function == "distance_matrix":
        coords_len = len(args['locations'])
        val = _distance_matrix_validation(args, coords_len)
    elif function == "pelias_search":
        val = _search_validation(args)
    elif function == "pelias_autocomplete":
        val = _autocomplete_validation(args)
    elif function == "pelias_structured":
        val = _structured_validation(args)
    elif function == "pelias_reverse":
        val = _reverse_validation(args)
    elif function == "pois":
        val = _pois_validation(args)
    elif function in ["elevation_point", "elevation_line"]:
        val = _elevation_validation(args)
        
    if val.errors:
        raise exceptions.ValidationError(val.errors)


def _directions_validation(params, coords_len):
    schema = {
        'coordinates': {'anyof': [{'type': ['list', 'tuple'], 'schema': {'type': 'float'}},
                                  {'type': ['list', 'tuple'],
                                   'schema': {'type': ['list', 'tuple'], 'schema': {'type': 'float'}}}],
                        'required': True},
        'profile': {'type': 'string',
                    'allowed': ['driving-car', 'driving-hgv', 'foot-walking', 'foot-hiking', 'cycling-regular',
                                'cycling-road', 'cycling-safe', 'cycling-mountain', 'cycling-tour',
                                'cycling-electric'], 'required': True},
        'preference': {'type': 'string', 'allowed': ['fastest', 'shortest', 'recommended'], 'default': 'fastest'},
        'format_out': {'type': 'string', 'allowed': ['json', 'geojson', 'gpx'], 'default': 'json'},
        'units': {'type': 'string', 'allowed': ['m', 'km', 'mi'], 'default': 'm'},
        'language': {'type': 'string',
                     'allowed': ['en', 'de', 'cn', 'es', 'ru', 'dk', 'fr', 'it', 'nl', 'br', 'se', 'tr', 'gr'],
                     'default': 'en'},
        'geometry': {'anyof': [
                {'type': 'string', 'allowed': ['true', 'false']},
                {'type': 'boolean'}
                ]},
        'geometry_format': {'type': 'string', 'allowed': ['encodedpolyline', 'geojson', 'polyline'],
                            'default': 'encodedpolyline'},
        'geometry_simplify': {'anyof': [
                {'type': 'string', 'allowed': ['true', 'false']},
                {'type': 'boolean'}
                ]},
        'instructions': {'anyof': [
                {'type': 'string', 'allowed': ['true', 'false']},
                {'type': 'boolean'}
                ]},
        'instructions_format': {'type': 'string', 'allowed': ['text', 'html'], 'default': 'text'},
        'roundabout_exits': {'anyof': [
                {'type': 'string', 'allowed': ['true', 'false']},
                {'type': 'boolean'}
                ]},
        'attributes': {'type': ['list', 'tuple'], 'schema': {'type': 'string',
                                                             'allowed': ['avgspeed',
                                                                         'detourfactor', 'percentage']}},
        'maneuvers': {'anyof': [
                {'type': 'string', 'allowed': ['true', 'false']},
                {'type': 'boolean'}
                ]},
        'radiuses': {'type': ['list', 'tuple'], 'schema': {'type': 'integer', 'min': -1},
                     'minlength': coords_len, 'maxlength': coords_len},
        'bearings': {'type': ['list', 'tuple'],
                     'minlength': coords_len - 1, 'maxlength': coords_len,
                     'schema': {'type': 'list',
                                'items': [{'type': 'integer', 'min': 0, 'max': 360, 'required': True},
                                          {'type': 'integer', 'default': 100}], 'minlength': 1, 'maxlength': 2},
                     },
        'continue_straight': {'anyof': [
                                    {'type': 'string', 'allowed': ['true', 'false']},
                                    {'type': 'boolean'}
                                    ],
                            'dependencies': {
                                '^profile': ['foot-walking', 'foot-hiking', 'cycling-regular',
                                            'cycling-road', 'cycling-safe', 'cycling-mountain',
                                            'cycling-tour',
                                            'cycling-electric']}
                                },
        'elevation': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]},
        'extra_info': {'type': ['list', 'tuple'], 'schema': {'type': 'string',
                                                             'allowed': ['steepness', 'suitability',
                                                                         'surface',
                                                                         'waycategory', 'waytype',
                                                                         'tollways',
                                                                         'traildifficulty']}},
        'optimized': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]},
        'options': {'type': 'dict', 'schema': {'maximum_speed': {'type': 'integer'},
                                               'avoid_features': {'type': 'list', 'schema': {'type': 'string',
                                                                                             'allowed': ['highways',
                                                                                                         'tollways',
                                                                                                         'ferries',
                                                                                                         'tunnels',
                                                                                                         'pavedroads',
                                                                                                         'unpavedroads',
                                                                                                         'tracks',
                                                                                                         'fords',
                                                                                                         'steps',
                                                                                                         'hills']}
                                                                  },
                                               'avoid_borders': {'type': 'string',
                                                                 'allowed': ['all', 'controlled'],
                                                                 'dependencies': {
                                                                     '^profile': ['driving-car', 'driving-hgv']}},
                                               'avoid_countries': {'schema': {'type': 'string'},
                                                                   'dependencies': {
                                                                       '^profile': ['driving-car', 'driving-hgv']}},
                                               'vehicle_type': {'type': 'string',
                                                                'allowed': ['hgv', 'bus', 'agricultural',
                                                                            'delivery',
                                                                            'forestry', 'goods'],
                                                                'dependencies': {'^profile': 'driving-hgv'}},
                                               'profile_params': {'type': 'dict', 'schema': {
                                                   'weightings': {'type': 'dict', 'schema': {
                                                       'steepness_difficulty': {'type': 'dict', 'schema': {
                                                           'level': {'type': 'integer', 'min': 0,
                                                                     'max': 3}},
                                                                                'dependencies': {
                                                                                    '^profile': ['cycling-regular',
                                                                                                 'cycling-road',
                                                                                                 'cycling-safe',
                                                                                                 'cycling-mountain',
                                                                                                 'cycling-tour',
                                                                                                 'cycling-electric']}},
                                                       'green': {'type': 'dict', 'schema': {
                                                           'factor': {'type': 'float', 'min': 0,
                                                                      'max': 1}},
                                                                 'dependencies': {'^profile': ['foot-walking',
                                                                                               'foot-hiking']}},
                                                       'quiet': {'type': 'dict', 'schema': {
                                                           'factor': {'type': 'float', 'min': 0,
                                                                      'max': 1}},
                                                                 'dependencies': {'^profile': ['foot-walking',
                                                                                               'foot-hiking']}}}},
                                                   'restrictions': {'type': 'dict', 'schema': {
                                                       'gradient': {'type': 'integer', 'min': 1, 'max': 15,
                                                                    'dependencies': {
                                                                        '^profile': ['cycling-regular',
                                                                                     'cycling-road',
                                                                                     'cycling-safe',
                                                                                     'cycling-mountain',
                                                                                     'cycling-tour',
                                                                                     'cycling-electric']}},
                                                       'length': {'type': 'integer',
                                                                  'dependencies': {  # '^vehicle_type',
                                                                      '^profile': 'driving-hgv'

                                                                  }},
                                                       'width': {'type': 'integer',
                                                                 'dependencies': {'^profile': 'driving-hgv',
                                                                                  # 'vehicle_type': True
                                                                                  }},
                                                       'height': {'type': 'integer',
                                                                  'dependencies': {'^profile': 'driving-hgv',
                                                                                   # 'vehicle_type': True
                                                                                   }},
                                                       'axleload': {'type': 'integer',
                                                                    'dependencies': {'^profile': 'driving-hgv',
                                                                                     # 'vehicle_type': True
                                                                                     }},
                                                       'weight': {'type': 'integer',
                                                                  'dependencies': {'^profile': 'driving-hgv',
                                                                                   # 'vehicle_type': True
                                                                                   }},
                                                       'hazmat': {'type': 'string', 'allowed': ['true', 'false'],
                                                                  'default': 'false',
                                                                  'dependencies': {'^profile': 'driving-hgv',
                                                                                   # 'vehicle_type': True
                                                                                   }},
                                                       # 'surface_type': {'type': 'string',
                                                       #                  'default': 'cobblestone:flattened',
                                                       #                  'dependencies': {'^profile': 'wheelchair'}},
                                                       # 'track_type': {'type': 'string', 'default': 'grade1',
                                                       #                'dependencies': {'^profile': 'wheelchair'}},
                                                       # 'smoothness_type': {'type': 'string', 'default': 'good',
                                                       #                     'dependencies': {'^profile': 'wheelchair'}},
                                                       # 'maximum_sloped_curb': {'anyof': [
                                                       #     {'type': 'float', 'allowed': [0.03, 0.06, 0.1],
                                                       #      'default': 0.06},
                                                       #     {'type': 'string', 'allowed': ['all']}], 'dependencies': {
                                                       #     '^profile': 'wheelchair'}},
                                                       # 'maximum_incline': {'anyof': [
                                                       #     {'type': 'integer', 'allowed': [3, 6, 10, 15],
                                                       #      'default': 6},
                                                       #     {'type': 'string', 'allowed': ['all']}], 'dependencies': {
                                                       #     '^profile': 'wheelchair'}}
                                                   }}
                                               }},
                                               'avoid_polygons': {}
                                               }},
        'id': {'type': 'string'},
        'dry_run': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]}
                        
    }

    v.validate(params, schema)

    return v


def _isochrones_validation(params):
    schema = {
        'locations': {'anyof': [{'type': ['list', 'tuple'], 'schema': {'type': 'float'}},
                                {'type': ['list', 'tuple'],
                                 'schema': {'type': ['list', 'tuple'], 'schema': {'type': 'float'}}}],
                      'required': True},
        'profile': {'type': 'string',
                    'allowed': ['driving-car', 'driving-hgv', 'foot-walking', 'foot-hiking', 'cycling-regular',
                                'cycling-road', 'cycling-safe', 'cycling-mountain', 'cycling-tour',
                                'cycling-electric'], 'default': 'driving-car', 'required': True},
        'range_type': {'type': 'string', 'allowed': ['time', 'distance'], 'default': 'time'},
        'intervals': {'type': ['list', 'tuple'], 'schema': {'type': 'integer'}, 'required': True},
        'segments': {'anyof': [{'type': ['list', 'tuple'], 'schema': {'type': 'integer'}},
                               {'type': 'integer'}]},
        'units': {'type': 'string', 'allowed': ['m', 'km', 'mi'], 'default': 'm'},
        'location_type': {'type': 'string', 'allowed': ['start', 'destination'], 'default': 'start'},
        'attributes': {'type': ['list', 'tuple'], 'schema': {'type': 'string',
                                                             'allowed': ['area', 'reachfactor', 'total_pop']}},
        'options': {'type': 'dict', 'schema': {'maximum_speed': {'type': 'integer'},
                                               'avoid_features': {'type': 'list', 'schema': {'type': 'string',
                                                                                             'allowed': ['highways',
                                                                                                         'tollways',
                                                                                                         'ferries',
                                                                                                         'tunnels',
                                                                                                         'pavedroads',
                                                                                                         'unpavedroads',
                                                                                                         'tracks',
                                                                                                         'fords',
                                                                                                         'steps',
                                                                                                         'hills']}},
                                               'avoid_borders': {'type': 'string',
                                                                 'allowed': ['all', 'controlled'],
                                                                 'dependencies': {
                                                                     '^profile': ['driving-car', 'driving-hgv']}},
                                               'avoid_countries': {'schema': {'type': 'string'},
                                                                   'dependencies': {
                                                                       '^profile': ['driving-car', 'driving-hgv']}},
                                               'vehicle_type': {'type': 'string',
                                                                'allowed': ['hgv', 'bus', 'agricultural',
                                                                            'delivery',
                                                                            'forestry', 'goods'],
                                                                'dependencies': {'^profile': 'driving-hgv'}},
                                               'profile_params': {'type': 'dict', 'schema': {
                                                   'weightings': {'type': 'dict', 'schema': {
                                                       'steepness_difficulty': {'type': 'dict', 'schema': {
                                                           'level': {'type': 'integer', 'min': 0,
                                                                     'max': 3}},
                                                                                'dependencies': {
                                                                                    '^profile': ['cycling-regular',
                                                                                                 'cycling-road',
                                                                                                 'cycling-safe',
                                                                                                 'cycling-mountain',
                                                                                                 'cycling-tour',
                                                                                                 'cycling-electric']}},
                                                       'green': {'type': 'dict', 'schema': {
                                                           'factor': {'type': 'float', 'min': 0,
                                                                      'max': 1}},
                                                                 'dependencies': {'^profile': ['foot-walking',
                                                                                               'foot-hiking']}},
                                                       'quiet': {'type': 'dict', 'schema': {
                                                           'factor': {'type': 'float', 'min': 0,
                                                                      'max': 1}},
                                                                 'dependencies': {'^profile': ['foot-walking',
                                                                                               'foot-hiking']}}}},
                                                   'restrictions': {'type': 'dict', 'schema': {
                                                       'gradient': {'type': 'integer', 'min': 1, 'max': 15,
                                                                    'dependencies': {
                                                                        '^profile': ['cycling-regular',
                                                                                     'cycling-road',
                                                                                     'cycling-safe',
                                                                                     'cycling-mountain',
                                                                                     'cycling-tour',
                                                                                     'cycling-electric']}},
                                                       'length': {'type': 'integer',
                                                                  'dependencies': {'^profile': 'driving-hgv',
                                                                                   # 'vehicle_type': True
                                                                                   }},
                                                       'width': {'type': 'integer',
                                                                 'dependencies': {'^profile': 'driving-hgv',
                                                                                  # 'vehicle_type': True
                                                                                  }},
                                                       'height': {'type': 'integer',
                                                                  'dependencies': {'^profile': 'driving-hgv',
                                                                                   # 'vehicle_type': True
                                                                                   }},
                                                       'axleload': {'type': 'integer',
                                                                    'dependencies': {'^profile': 'driving-hgv',
                                                                                     # 'vehicle_type': True
                                                                                     }},
                                                       'weight': {'type': 'integer',
                                                                  'dependencies': {'^profile': 'driving-hgv',
                                                                                   # 'vehicle_type': True
                                                                                   }},
                                                       'hazmat': {'type': 'string', 'allowed': ['true', 'false'],
                                                                  'default': 'false',
                                                                  'dependencies': {'^profile': 'driving-hgv',
                                                                                   # 'vehicle_type': True
                                                                                   }},
                                                       # 'surface_type': {'type': 'string',
                                                       #                  'default': 'cobblestone:flattened',
                                                       #                  'dependencies': {'^profile': 'wheelchair'}},
                                                       # 'track_type': {'type': 'string', 'default': 'grade1',
                                                       #                'dependencies': {'^profile': 'wheelchair'}},
                                                       # 'smoothness_type': {'type': 'string', 'default': 'good',
                                                       #                     'dependencies': {'^profile': 'wheelchair'}},
                                                       # 'maximum_sloped_curb': {'anyof': [
                                                       #     {'type': 'float', 'allowed': [0.03, 0.06, 0.1],
                                                       #      'default': 0.06},
                                                       #     {'type': 'string', 'allowed': ['all']}], 'dependencies': {
                                                       #     '^profile': 'wheelchair'}},
                                                       # 'maximum_incline': {'anyof': [
                                                       #     {'type': 'integer', 'allowed': [3, 6, 10, 15],
                                                       #      'default': 6},
                                                       #     {'type': 'string', 'allowed': ['all']}], 'dependencies': {
                                                       #     '^profile': 'wheelchair'}}
                                                   }}
                                               }},
                                               'avoid_polygons': {}
                                               }},
        'intersections': {'anyof': [
                            {'type': 'string', 'allowed': ['true', 'false']},
                            {'type': 'boolean'}
                            ]},
        'id': {'type': 'string'},
        'smoothing': {"type": "float", 'min': 0, 'max': 1},
        'dry_run': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]}
    }

    v.validate(params, schema)

    return v


def _distance_matrix_validation(params, coords_len):
    schema = {
        'locations': {'anyof': [{'type': ['list', 'tuple'], 'schema': {'type': 'float'}},
                                {'type': ['list', 'tuple'],
                                 'schema': {'type': ['list', 'tuple'], 'schema': {'type': 'float'}}}],
                      'required': True},
        'profile': {'type': 'string',
                    'allowed': ['driving-car', 'driving-hgv', 'foot-walking', 'foot-hiking', 'cycling-regular',
                                'cycling-road', 'cycling-safe', 'cycling-mountain', 'cycling-tour',
                                'cycling-electric'], 'default': 'driving-car', 'required': True},
        'sources': {'anyof': [
            {'type': 'list', 'schema': {'type': 'integer', 'min': 0, 'max': coords_len - 1}},
            {'type': 'string', 'allowed': ['all']}]},
        'destinations': {'anyof': [
            {'type': 'list', 'schema': {'type': 'integer', 'min': 0, 'max': coords_len - 1}},
            {'type': 'string', 'allowed': ['all']}]},
        'metrics': {'type': 'list', 'schema': {'type': 'string'}, 'allowed': ['distance', 'duration']},
        'resolve_locations': {'anyof': [
                                    {'type': 'string', 'allowed': ['true', 'false']},
                                    {'type': 'boolean'}
                                    ]},
        'units': {'type': 'string', 'allowed': ['m', 'km', 'mi'], 'default': 'm'},
        'optimized': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]},
        'id': {'type': 'string'},
        'dry_run': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]}
    }

    v.validate(params, schema)

    return v


def _search_validation(params):
    schema = {
        'text': {'type': 'string', 'required': True},
        'focus_point': {'type': 'list', 'schema': {'type': 'float'}},
        'rect_min_x': {'type': 'float'},
        'rect_min_y': {'type': 'float'},
        'rect_max_x': {'type': 'float'},
        'rect_max_y': {'type': 'float'},
        'circle_point': {'type': 'list', 'schema': {'type': 'float'}},
        'circle_radius': {'type': 'integer', 'default': 50},
        'sources': {'type': 'list', 'schema': {'type': 'string'}, 'allowed': ['osm', 'oa', 'wof', 'gn'],
                    'default': ['osm', 'oa', 'wof', 'gn']},
        'layers': {'type': 'list', 'schema': {'type': 'string'},
                   'allowed': ['venue', 'address', 'street', 'neighbourhood', 'borough', 'localadmin', 'locality',
                               'county', 'macrocounty', 'region', 'macroregion', 'country', 'coarse'],
                   'default': ['venue', 'address', 'street', 'neighbourhood', 'borough', 'localadmin', 'locality',
                               'county', 'macrocounty', 'region', 'macroregion', 'country', 'coarse']},
        'country': {'type': 'string'},
        'size': {'type': 'integer', 'default': 10},
        'dry_run': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]}
    }

    v.validate(params, schema)

    return v


def _autocomplete_validation(params):
    schema = {
        'text': {'type': 'string', 'required': True},
        'focus_point': {'type': 'list', 'schema': {'type': 'float'}},
        'rect_min_x': {'type': 'float'},
        'rect_min_y': {'type': 'float'},
        'rect_max_x': {'type': 'float'},
        'rect_max_y': {'type': 'float'},
        'sources': {'type': 'list', 'schema': {'type': 'string'}, 'allowed': ['osm', 'oa', 'wof', 'gn'],
                    'default': ['osm', 'oa', 'wof', 'gn']},
        'layers': {'type': 'list', 'schema': {'type': 'string'},
                   'allowed': ['venue', 'address', 'street', 'neighbourhood', 'borough', 'localadmin', 'locality',
                               'county', 'macrocounty', 'region', 'macroregion', 'country', 'coarse'],
                   'default': ['venue', 'address', 'street', 'neighbourhood', 'borough', 'localadmin', 'locality',
                               'county', 'macrocounty', 'region', 'macroregion', 'country', 'coarse']},
        'country': {'type': 'string'},
        'dry_run': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]}
    }

    v.validate(params, schema)

    return v


def _structured_validation(params):
    schema = {
        'address': {'type': 'string'},
        'neighbourhood': {'type': 'string'},
        'borough': {'type': 'string'},
        'locality': {'type': 'string'},
        'county': {'type': 'string'},
        'region': {'type': 'string'},
        'postalcode': {'type': 'integer'},
        'country': {'type': 'string'},
        'dry_run': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]}
    }

    v.validate(params, schema)

    return v


def _reverse_validation(params):
    schema = {
        'point': {'type': ['list', 'tuple'], 'schema': {'type': 'float'}, 'required': True},
        'circle_radius': {'type': 'integer'},
        'sources': {'type': ['list', 'tuple'], 'schema': {'type': 'string'}, 'allowed': ['osm', 'oa', 'wof', 'gn'],
                    'default': ['osm', 'oa', 'wof', 'gn']},
        'layers': {'type': ['list', 'tuple'], 'schema': {'type': 'string'},
                   'allowed': ['venue', 'address', 'street', 'neighbourhood', 'borough', 'localadmin', 'locality',
                               'county', 'macrocounty', 'region', 'macroregion', 'country', 'coarse'],
                   'default': ['venue', 'address', 'street', 'neighbourhood', 'borough', 'localadmin', 'locality',
                               'county', 'macrocounty', 'region', 'macroregion', 'country', 'coarse']},
        'country': {'type': 'string'},
        'size': {'type': 'integer', 'default': 10},
        'dry_run': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]}
    }

    v.validate(params, schema)

    return v


def _pois_validation(params):
    schema = {
        'request': {'type': 'string', 'allowed': ['pois', 'list', 'stats'], 'default': 'pois', 'required': True},
        # 'geometry': {'type': 'dict', 'schema': {
        'bbox': {'type': 'list', 'schema': {'type': 'list', 'schema': {'type': 'float', 'maxlength': 2}}},
        'geojson': {'type': 'dict', 'schema': {
            'type': {'type': 'string', 'allowed': ['Point', 'Polygon', 'LineString'], 'required': True,
                     'dependencies': 'coordinates'},
            'coordinates': {'anyof': [
                {'type': 'list', 'schema': {'type': 'float', 'maxlength': 2}, 'dependencies': {'type': 'Point'}},
                {'type': 'list',
                 'schema': {'type': 'list', 'schema': {'type': 'float', 'maxlength': 2}}, 'dependencies': {'type': ['LineString']}},
                {'type': 'list',
                 'schema': {'type': 'list',
                            'schema': {'type': 'list', 'schema': {'type': 'float', 'maxlength': 2}}},
                 'dependencies': {'type': ['Polygon']}}]
            }, }},
        'buffer': {'type': 'integer', 'default': 500},
        # }},
        # 'filters': {'type': 'dict', 'schema': {
        'filter_category_group_ids': {'type': 'list', 'schema': {'type': 'integer'}},
        'filter_category_ids': {'type': 'list', 'schema': {'type': 'integer'}},
        'filters_custom': {'type': 'dict', 'schema': {
            'wheelchair': {'type': 'list',
                           'schema': {'type': 'string', 'allowed': ['yes', 'no', 'limited', 'designated']}},
            'smoking': {'type': 'list', 'schema': {'type': 'string',
                                                   'allowed': ['dedicated', 'yes', 'no', 'separated', 'isolated',
                                                               'outside']}},
            'fee': {'type': 'list', 'schema': {'type': 'string', 'allowed': ['yes', 'no']}},
            'name': {'type': 'string'},}},
        # }},
        'limit': {'type': 'integer', 'max': 1000},
        'sortby': {'type': 'string', 'allowed': ['category', 'distance']},
        'dry_run': {'anyof': [
                        {'type': 'string', 'allowed': ['true', 'false']},
                        {'type': 'boolean'}
                        ]}
    }

    v.validate(params, schema)

    return v

def _elevation_validation(params):
    schema = {
            'format_in': {'type': 'string', 'allowed': ['geojson', 'point', 'polyline', 'encodedpolyline'], 'required': True},
            'format_out': {'type': 'string', 'allowed': ['geojson', 'point', 'polyline', 'encodedpolyline'], 'default': 'geojson'},
            'dataset': {'type': 'string', 'allowed': ['srtm'], 'default': 'srtm'},
            'geometry': {'anyof': [
                    {'type': 'dict', 'schema':
                        {'type': {'type': 'string', 'allowed': ['Point', 'LineString'], 'required': True},
                        'coordinates': {
                                'anyof': [
                                    {'type': 'list', 
                                     'schema': {'type': 'float', 'minlength': 2, 'maxlength': 2}},                       
                                    {'type': 'list',
                                     'schema': {'type': 'list', 'schema': {'type': 'float',
                                                                           'minlength': 2,
                                                                           'maxlength': 2}}}
                                    ],
                                'required': True
                                }
                        },
                    'dependencies': {'^format_in': 'geojson'},
                    },
                    {'type': 'list',
                     'schema': {'type': 'float', 'minlength': 2, 'maxlength': 2},
                     'dependencies': {'^format_in': 'point'}
                     },
                    {'type': ['list', 'tuple'],
                     'schema': {'type': ['list', 'tuple'],
                              'schema': {'type': 'float', 'minlength': 2, 'maxlength': 2}},
                     'dependencies': {'^format_in': 'polyline'}
                     },
                    {'type': 'string',
                     'dependencies': {'^format_in': 'encodedpolyline'}
                     }
                     ],
                    'required': True},
            'dry_run': {'anyof': [
                            {'type': 'string', 'allowed': ['true', 'false']},
                            {'type': 'boolean'}
                            ]}
                }
    v.validate(params, schema)
    
    return v


# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All rights reserved.
#
# Modifications Copyright (C) 2018 HeiGIT, University of Heidelberg.
#
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
#

"""Tests for the directions module."""

import responses

import openrouteservice
from openrouteservice.exceptions import ValidationError
import test as _test

class DirectionsTest(_test.TestCase):
            
    def setUp(self):
        self.key = 'sample_key'
        self.client = openrouteservice.Client(self.key)
        self.coords_valid = ((8.34234,48.23424),(8.34423,48.26424))

    @responses.activate
    def test_simple_directions(self):
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/directions',
                      body='{"status":"OK","routes":[]}',
                      status=200,
                      content_type='application/json')
        
        routes = self.client.directions(self.coords_valid)

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.openrouteservice.org/directions?'
                            'api_key={}&profile=driving-car&'
                            'coordinates=8.34234,48.23424|8.34423,48.26424&'
                            'optimized=true'.format(self.key),
                            responses.calls[0].request.url)        

    @responses.activate
    def test_bearings(self):
        # First: test correct output for bearings 
        # Second: should switch to optimized=false bcs bearings was specified
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/directions',
                      body='{"status":"OK","routes":[]}',
                      status=200,
                      content_type='application/json')
        
        # Simplest directions request. Driving directions by default.
        routes = self.client.directions(self.coords_valid,
                                        bearings=[[100,100],[200,200]])

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.openrouteservice.org/directions?'
                            'api_key={}&coordinates=8.34234%2C48.23424%7C8.34423%2C48.26424&'
                            'profile=driving-car&bearings=100%2C100%7C200%2C200&optimized=false'.format(self.key),
                            responses.calls[0].request.url)

    @responses.activate
    def test_continue_straight(self):
        # Should switch to optimized=false bcs continue_straight is set true
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/directions',
                      body='{"status":"OK","routes":[]}',
                      status=200,
                      content_type='application/json')
        
        # Simplest directions request. Driving directions by default.
        routes = self.client.directions(self.coords_valid,
                                        profile='cycling-regular',
                                        continue_straight='true')

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.openrouteservice.org/directions?'
                            'api_key={}&coordinates=8.34234%2C48.23424%7C8.34423%2C48.26424&'
                            'profile=cycling-regular&continue_straight=true&optimized=false'.format(self.key),
                            responses.calls[0].request.url)        

    @responses.activate
    def test_complex_request(self):
        responses.add(responses.GET,
                      'https://api.openrouteservice.org/directions',
                      body='{"status":"OK","routes":[]}',
                      status=200,
                      content_type='application/json')
        
        params = {
                'coordinates':self.coords_valid,
                'profile':'cycling-regular',
                'preference':'fastest',
                'units':'mi',
                'language':'en',
                'geometry':'true',
                'geometry_format':'geojson',
                'geometry_simplify':'false',
                'instructions':'false',
                'instructions_format':'html',
                'roundabout_exits':'true',
                'attributes':['avgspeed'],
                'radiuses':[10000,10000],
                'bearings':[[100,100], [200,200]],
                'continue_straight':'false',
                'elevation':'true',
                'extra_info':['steepness', 'suitability'],
                'optimized':'false',
                'options':{'maximum_speed':20}
                }

        print(params)
        routes = self.client.directions(**params)
                                        

        self.assertEqual(1, len(responses.calls))
        self.assertURLEqual('https://api.openrouteservice.org/directions?api_key={}&'
                            'coordinates=8.34234%2C48.23424%7C8.34423%2C48.26424&'
                            'profile=cycling-regular&preference=fastest&units=mi&'
                            'language=en&geometry=true&geometry_format=geojson&'
                            'geometry_simplify=false&instructions=false&'
                            'instructions_format=html&roundabout_exits=true'
                            '&attributes=avgspeed'
                            '&radiuses=10000%7C10000&bearings=100%2C100%7C200%2C200&'
                            'continue_straight=false&elevation=true&'
                            'extra_info=steepness%7Csuitability&optimized=false&'
                            'options=%7B%22maximum_speed%22%3A+20%7D'.format(self.key),
                            responses.calls[0].request.url)
